from typing import cast as typing_cast, Any, TypeVar, ParamSpec
from collections.abc import Callable, Sized, Sequence
from functools import wraps, singledispatch

from mpi4py import MPI
import numpy as np

from .._independant_mpi._mpi_config import mpi_config as MPI_Config
from ...Tools._edit_locals import use_locals



P = ParamSpec("P")
T = TypeVar("T")
_MAX_BUFFER_SIZE = 2**31 - 1 # Max length of an signed int32



@use_locals(0, 0)
def synchronyse(target: T, /, root: int|None = None, comm: MPI.Intracomm|None = None) -> T:
    """
    Broadcast the value of a variable on the root rank to the same variable on all ranks in the communicator.

    The first argument should be a string containing the variable name NOT the value of the variable!
    """
    if root == None:
        root = MPI_Config.root
    if comm == None:
        comm = MPI_Config.comm
    result = comm.bcast(target, root)
    comm.barrier()
    return result



def mpi_check_equal(value: T, /, root: int|None = None, comm: MPI.Intracomm|None = None) -> bool:
    """
    Check all ranks have the same value.
    """
    comm = MPI_Config.allow_default_comm(comm)
    root = MPI_Config.allow_default_root(root)
    are_equal: bool
    values: list[T]|None = comm.gather(value, root = root)
    if comm.rank == root:
        are_equal = True
        for v in values[1:]:
            are_equal = values[0] == v
            if not are_equal:
                break
    synchronyse("are_equal", root = root, comm = comm)
    return are_equal



def if_mpi_root(default: T, root: int|None = None):
    """
    Function decorator for restricting calls to the root rank.

    A default return value of the appropriate return type must be provided (this can be None if the wrapped function allows it or specifies no return type).
    """
    if root == None:
        root = MPI_Config.root
    def inner(func: Callable[P, T]) -> Callable[P, T]:
        @wraps
        def wrapper(*args: P.args, **kwargs: P.kwargs):
            if MPI_Config.rank == root:
                return func(*args, **kwargs)
            else:
                return default
        return wrapper
    return inner



def mpi_barrier(comm: MPI.Intracomm|None = None) -> None:
    """
    Wait for all ranks in the communicator to reach this barrier.
    """
    (comm if comm is not None else MPI_Config.comm).barrier()



def mpi_sum(data: Sequence[T], comm: MPI.Intracomm|None = None, root: int|None = None) -> T:
    """
    Calculate the sum of data across multiple ranks.
    """

    comm = MPI_Config.allow_default_comm(comm)
    root = MPI_Config.allow_default_root(root)

    comm.barrier()

    local_sum: T|None = sum(data[1:], start = data[0]) if len(data) > 1 else data[0] if len(data) > 0 else None

    result: T
    has_valid_data: bool
    rank_sums: list[T|None]|None = comm.gather(local_sum, root = root)
    if MPI_Config.check_is_root(root = root):
        valid_data: list[T] = [v for v in typing_cast(list[T], rank_sums) if v is not None]
        has_valid_data = len(valid_data) > 0
        if has_valid_data:
            result = sum(valid_data[1:], start = valid_data[0]) if len(valid_data) > 1 else valid_data[0]

    synchronyse("has_valid_data", root = root, comm = comm)
    if not has_valid_data:
        raise IndexError("No data provided by any rank.")

    synchronyse("result", root = root, comm = comm)

    comm.barrier()

    return result



def mpi_mean(data: Sequence[float], weights: Sequence[float]|None = None, comm: MPI.Intracomm|None = None, root: int|None = None) -> float:
    """
    Calculate the mean of data across multiple ranks.
    Optionally, specify weights to calculate the weighted mean.
    """

    comm = MPI_Config.allow_default_comm(comm)
    root = MPI_Config.allow_default_root(root)

    comm.barrier()

    local_sum: float = np.sum(data) if weights is None else np.sum(np.array(data) * np.array(weights))
    local_summed_divisor: float = len(data) if weights is None else np.sum(weights)

    result: float
    divide_by_zero: bool
    rank_sums: list[float]|None = comm.gather(local_sum, root = root)
    rank_divisor_sums: list[float]|None = comm.gather(local_summed_divisor, root = root)
    if MPI_Config.check_is_root(root = root):
        divisor = sum(typing_cast(list[float], rank_divisor_sums))
        divide_by_zero = divisor == 0
        if not divide_by_zero:
            result = sum(typing_cast(list[float], rank_sums)) / divisor

    synchronyse("divide_by_zero", root = root, comm = comm)
    if divide_by_zero:
        if weights is None:
            raise IndexError("No data provided by any rank.")
        else:
            raise ZeroDivisionError("Sum of weights was 0.")

    synchronyse("result", root = root, comm = comm)

    comm.barrier()

    return result



def mpi_get_slice(length: int|Sized, comm: MPI.Intracomm|None = None, rank: int|None = None) -> slice:
    """
    Partition a set of data such that it is distributed evenly accross each MPI rank.

    This method can be used to help get data from a h5py dataset when using parallel IO.

    Parameters:
                 int|Sized length -> The length of the target data or the target data if a size can be derived from it.
        MPI.Intracomm|None comm   -> Optional MPI communicator object (defaults to the one from MPI_Config).
                  int|None comm   -> Optional MPI rank (defaults to the one from MPI_Config if not comm is specified or the one from the provided comm).

    Returns:
        slice -> A slice object that can be applied a Sequence of the appropriate length.
    """

    comm_size = MPI_Config.comm_size if comm is None else comm.Get_size()
    if rank == None:
        rank = MPI_Config.rank if comm is None else comm.Get_rank()

    if not isinstance(length, int):
        length = len(length)

    base_chunk_size = length // comm_size
    chunks_with_one_extra = length % comm_size
    chunk_sizes = np.full(comm_size, base_chunk_size, dtype = int)
    chunk_sizes[:chunks_with_one_extra] += 1
    snaphot_data_offset: int = chunk_sizes[:rank].sum()
    snaphot_data_chunk_size: int = chunk_sizes[rank]

    return slice(snaphot_data_offset, snaphot_data_offset + snaphot_data_chunk_size)



def mpi_slice(data: Sequence[T], comm: MPI.Intracomm|None = None, rank: int|None = None) -> Sequence[T]:
    """
    Partition a set of data such that it is distributed evenly accross each MPI rank.

    This method can be used to quickly get data from a h5py dataset when using parallel IO.

    Parameters:
                  Sequence data -> Data to be partitioned.
        MPI.Intracomm|None comm -> Optional MPI communicator object (defaults to the one from MPI_Config).
                  int|None comm -> Optional MPI rank (defaults to the one from MPI_Config if not comm is specified or the one from the provided comm).

    Returns:
        Sequence -> Section of the data assigned to the calling rank.
    """
    return data[mpi_get_slice(len(data), comm, rank)]



def mpi_gather_array(data: np.ndarray, comm: MPI.Intracomm|None = None, root: int|None = None, target_buffer: np.ndarray|None = None, return_chunk_sizes: bool = False, allgather: bool = False) -> np.ndarray|None|tuple[np.ndarray, np.ndarray]|tuple[None, None]:
    """
    Gather numpy array data to the root rank.
    If the resulting array is larger than the MPI buffer length, data will be transmitted point-to-point in rank order instead of using the Gatherv method.
    """
    comm = MPI_Config.allow_default_comm(comm)
    root = MPI_Config.allow_default_root(root)

    comm.barrier()

    if sum(comm.allgather(int(allgather))) not in (0, comm.size):
        raise ValueError("Not all ranks set parameter \"allgather\" as True.")
    comm.barrier()

    if not allgather:
        if any(comm.allgather(comm.rank != root and target_buffer is not None)):
            raise TypeError("Output buffer provided by non-root rank.")
        comm.barrier()

    local_buffer_length = np.prod(data.shape)
    local_buffer_length_first_dimension = data.shape[0]
    one_dimension = local_buffer_length == local_buffer_length_first_dimension
    local_buffer_step_size = 1 if one_dimension else np.prod(data.shape[1:])

    if not mpi_check_equal([one_dimension, local_buffer_step_size], root = root, comm = comm):
        raise BufferError("Input buffers on ranks have different shapes beyond first dimension.")
    comm.barrier()

    use_manual_transfer: int
    input_buffer_lengths_first_dimension: list[int] = comm.allgather(data.shape[0], root = root)
    comm.barrier()
    if comm.rank == root or allgather:
        input_buffer_lengths_first_dimension: np.ndarray = np.array(input_buffer_lengths_first_dimension, dtype = int) # Why is this necessary???
        output_buffer_length_first_dimension: int = sum(input_buffer_lengths_first_dimension)
        output_buffer_length = output_buffer_length_first_dimension * local_buffer_step_size
        use_manual_transfer = output_buffer_length > _MAX_BUFFER_SIZE
    synchronyse("use_manual_transfer", root = root, comm = comm)

    can_continue: bool
    if comm.rank == root:
        can_continue = target_buffer is None or target_buffer.shape == (output_buffer_length_first_dimension, *data.shape[1:])
    synchronyse("can_continue", root = root, comm = comm)
    if not can_continue:
        raise BufferError("Input buffer size/shape is not correct for the input data.")
    
    if comm.rank == root or allgather:
        if target_buffer is None:
            target_buffer = np.empty(shape = (output_buffer_length_first_dimension, *data.shape[1:]), dtype = data.dtype)
        rank_offsets_first_dimension = np.insert(np.cumsum(input_buffer_lengths_first_dimension), 0, 0)[:-1]
        rank_offsets = np.insert(np.cumsum(input_buffer_lengths_first_dimension * local_buffer_step_size), 0, 0)[:-1]

    # The output buffer is within the maximum allowed size
    if not use_manual_transfer:
        comm.barrier()
        if allgather:
            comm.Allgatherv(data, None if comm.rank != root else (target_buffer, (input_buffer_lengths_first_dimension * local_buffer_step_size, rank_offsets)))
        else:
            comm.Gatherv(data, None if comm.rank != root else (target_buffer, (input_buffer_lengths_first_dimension * local_buffer_step_size, rank_offsets)), root = root)
        comm.barrier()

    # The output buffer is larger than the maximum buffer length
    # Data must be communicated manualy
    else:
        if comm.rank == root or allgather:
            target_buffer[rank_offsets_first_dimension[comm.rank]:data.shape[0]] = data[:]
        comm.barrier()
        for i in range(0, comm.size):
            if not allgather and i == root:
                # The root has already had an opportunity to transfer its data so we can skip this one
                continue
            if comm.rank != i:
                total_expected_elements_all_dimensions = input_buffer_lengths_first_dimension[i] * local_buffer_step_size
                if total_expected_elements_all_dimensions <= _MAX_BUFFER_SIZE:
                    # Data can be transferred in a single function call
                    if allgather:
                        comm.Bcast(target_buffer[rank_offsets_first_dimension[i] : rank_offsets_first_dimension[i] + input_buffer_lengths_first_dimension[i]], root = i)
                    else:
                        comm.Recv(target_buffer[rank_offsets_first_dimension[i] : rank_offsets_first_dimension[i] + input_buffer_lengths_first_dimension[i]], source = i)
                else:
                    # Data is too large be transferred in a single function call
                    # Multiple calls required
                    local_chunk_offset: int = 0
                    elements_first_dimension_remaining: int = input_buffer_lengths_first_dimension[i]
                    while elements_first_dimension_remaining > 0:
                        chunk_size_this_transfer = _MAX_BUFFER_SIZE if elements_first_dimension_remaining > _MAX_BUFFER_SIZE else elements_first_dimension_remaining
                        if allgather:
                            comm.Bcast(target_buffer[rank_offsets_first_dimension[i] + local_chunk_offset : rank_offsets_first_dimension[i] + local_chunk_offset + chunk_size_this_transfer], root = i)
                        else:
                            comm.Recv(target_buffer[rank_offsets_first_dimension[i] + local_chunk_offset : rank_offsets_first_dimension[i] + local_chunk_offset + chunk_size_this_transfer], source = i)
                        local_chunk_offset += chunk_size_this_transfer
                        elements_first_dimension_remaining -= chunk_size_this_transfer
            elif comm.rank == i:
                if np.prod(data.shape) <= _MAX_BUFFER_SIZE:
                    if allgather:
                        comm.Bcast(data, root = comm.rank)
                    else:
                        comm.Send(data, dest = root)
                else:
                    local_chunk_offset: int = 0
                    elements_first_dimension_remaining: int = data.shape[0]
                    while elements_first_dimension_remaining > 0:
                        chunk_size_this_transfer = _MAX_BUFFER_SIZE if elements_first_dimension_remaining > _MAX_BUFFER_SIZE else elements_first_dimension_remaining
                        if allgather:
                            comm.Bcast(data[local_chunk_offset : local_chunk_offset + chunk_size_this_transfer], root = comm.rank)
                        else:
                            comm.Send(data[local_chunk_offset : local_chunk_offset + chunk_size_this_transfer], dest = root)
                        local_chunk_offset += chunk_size_this_transfer
                        elements_first_dimension_remaining -= chunk_size_this_transfer
            comm.barrier()

    if return_chunk_sizes:
        return target_buffer, input_buffer_lengths_first_dimension
    else:
        return target_buffer



def mpi_scatter_array(data: np.ndarray|None, elements_this_rank: int|None = None, elements_per_rank: list[int]|None = None, comm: MPI.Intracomm|None = None, root: int|None = None, target_buffer: np.ndarray|None = None) -> np.ndarray:
    """
    Scatter numpy array data from the root rank to all ranks.
    If the target array is larger than the MPI buffer length, data will be transmitted point-to-point in rank order instead of using the Scatterv method.
    """
    can_continue: bool # This is used to synchronise error conditions between ranks

    comm = MPI_Config.allow_default_comm(comm)
    root = MPI_Config.allow_default_root(root)

    comm.barrier()

    if comm.rank == root:
        can_continue = data is not None
    synchronyse("can_continue", root = root, comm = comm)
    if not can_continue:
        raise TypeError("Root rank did not provide any data.")

    if any(comm.allgather(comm.rank != root and data is not None)):
        raise TypeError("Non-root rank provided input data.")
    comm.barrier()

    if any(comm.allgather(comm.rank != root and elements_per_rank is not None)):
        raise TypeError("Non-root rank provided scattering information for all ranks.")
    comm.barrier()

    does_rank_give_chunk_size = comm.allgather(elements_this_rank is not None)
    if any(does_rank_give_chunk_size) and not all(does_rank_give_chunk_size):
        raise TypeError("Some but not all ranks provided chunk size information.")
    comm.barrier()

    if comm.rank == root:
        can_continue = elements_this_rank is None or not any(does_rank_give_chunk_size)
    synchronyse("can_continue", root = root, comm = comm)
    if not can_continue:
        raise TypeError("Root rank provided chunk sizes but ranks also provided chunk sizes.")

    output_buffer_lengths_first_dimension: np.ndarray
    rank_offsets_first_dimension: np.ndarray
    if elements_this_rank is None:
        if comm.rank == root:
            if elements_per_rank is None:
                output_buffer_lengths_first_dimension = np.empty(comm.size, dtype = int)
                rank_offsets_first_dimension = np.empty(comm.size, dtype = int)
                for i in range(comm.size):
                    rank_chunk_start, rank_chunk_stop, _ = mpi_get_slice(data.shape[0], comm = comm, rank = i).indices(data.shape[0])
                    output_buffer_lengths_first_dimension[i] = rank_chunk_stop - rank_chunk_start
                    rank_offsets_first_dimension[i] = rank_chunk_start
            else:
                output_buffer_lengths_first_dimension = np.array(elements_per_rank, dtype = int)
                rank_offsets_first_dimension = np.insert(np.cumsum(output_buffer_lengths_first_dimension), 0, 0)[:-1]
        elements_this_rank = comm.scatter(None if comm.rank != root else output_buffer_lengths_first_dimension, root = root)
    else:
        output_buffer_lengths_first_dimension = comm.gather(elements_this_rank, root = root)
        comm.barrier()
        if comm.rank == root:
            output_buffer_lengths_first_dimension = np.array(output_buffer_lengths_first_dimension, dtype = int)
            rank_offsets_first_dimension = np.insert(np.cumsum(output_buffer_lengths_first_dimension), 0, 0)[:-1]

    local_buffer_shape_after_first_dimension: tuple[int, ...]
    if comm.rank == root:
        source_buffer_length = np.prod(data.shape)
        source_buffer_length_first_dimension = data.shape[0]
        one_dimension = source_buffer_length == source_buffer_length_first_dimension
        buffer_step_size = 1 if one_dimension else np.prod(data.shape[1:])
        local_buffer_shape_after_first_dimension = data.shape[1:]
    synchronyse("local_buffer_shape_after_first_dimension", root = root, comm = comm)

    if target_buffer is not None:
        if any(comm.allgather(target_buffer is not None and (target_buffer.shape[0] != elements_this_rank or target_buffer.shape[1:] != local_buffer_shape_after_first_dimension))):
            raise BufferError("Output buffers provided to ranks did not match their respective expected sizes/shapes.")
        comm.barrier()

    target_datatype: object
    target_datatype: object
    if comm.rank == root:
        target_datatype = data.dtype
    synchronyse("target_datatype", root = root, comm = comm)
    if target_buffer is None:
        target_buffer = np.empty(shape = (elements_this_rank, *local_buffer_shape_after_first_dimension), dtype = target_datatype)

    if comm.rank == root:
        rank_offsets = np.insert(np.cumsum(output_buffer_lengths_first_dimension * buffer_step_size), 0, 0)[:-1]

    use_manual_transfer: bool
    if comm.rank == root:
        use_manual_transfer = data.shape[0] > _MAX_BUFFER_SIZE
    synchronyse("use_manual_transfer", root = root, comm = comm)

    # The input buffer is within the maximum allowed size
    if not use_manual_transfer:
        comm.barrier()
        comm.Scatterv(None if comm.rank != root else (data, (output_buffer_lengths_first_dimension * buffer_step_size, rank_offsets)), target_buffer, root = root)
        comm.barrier()

    # The input buffer is larger than the maximum buffer length
    # Data must be communicated manualy
    else:
        if comm.rank == root:
            target_buffer[:] = data[:output_buffer_lengths_first_dimension[0]]
        comm.barrier()
        for i in range(1, comm.size):
            if comm.rank == root:
                total_expected_elements_all_dimensions = output_buffer_lengths_first_dimension[i] * buffer_step_size
                if total_expected_elements_all_dimensions <= _MAX_BUFFER_SIZE:
                    # Data can be transferred in a single function call
                    comm.Send(data[rank_offsets_first_dimension[i] : rank_offsets_first_dimension[i] + output_buffer_lengths_first_dimension[i] ], dest = i)
                else:
                    # Data is too large be transferred in a single function call
                    # Multiple calls required
                    local_chunk_offset: int = 0
                    elements_first_dimension_remaining: int = output_buffer_lengths_first_dimension[i]
                    while elements_first_dimension_remaining > 0:
                        chunk_size_this_transfer = _MAX_BUFFER_SIZE if elements_first_dimension_remaining > _MAX_BUFFER_SIZE else elements_first_dimension_remaining
                        comm.Send(data[rank_offsets_first_dimension[i] + local_chunk_offset : rank_offsets_first_dimension[i] + local_chunk_offset + chunk_size_this_transfer], dest = i)
                        local_chunk_offset += chunk_size_this_transfer
                        elements_first_dimension_remaining -= chunk_size_this_transfer
            elif comm.rank == i:
                if elements_this_rank * np.prod(local_buffer_shape_after_first_dimension) <= _MAX_BUFFER_SIZE:
                    comm.Recv(target_buffer, source = root)
                else:
                    local_chunk_offset: int = 0
                    elements_first_dimension_remaining: int = elements_this_rank
                    while elements_first_dimension_remaining > 0:
                        chunk_size_this_transfer = _MAX_BUFFER_SIZE if elements_first_dimension_remaining > _MAX_BUFFER_SIZE else elements_first_dimension_remaining
                        comm.Recv(target_buffer[local_chunk_offset : local_chunk_offset + chunk_size_this_transfer], source = root)
                        local_chunk_offset += chunk_size_this_transfer
                        elements_first_dimension_remaining -= chunk_size_this_transfer
            comm.barrier()

    return target_buffer



def mpi_redistribute_array_evenly(data: np.ndarray, elements_this_rank: int|None = None, elements_per_rank: list[int]|None = None, comm: MPI.Intracomm|None = None, root: int|None = None, target_buffer: np.ndarray|None = None) -> np.ndarray:
    """
    Send all data to the root rank for redistribution to all ranks.
    See mpi_gather_array for information regarting the "data" parameter.
    See mpi_scatter_array for information regarting the "elements_this_rank" and "elements_per_rank" parameters.
    """
#    comm = MPI_Config.allow_default_comm(comm)
#    root = MPI_Config.allow_default_root(root)
    return mpi_scatter_array(mpi_gather_array(data, comm = comm, root = root), target_buffer = target_buffer, elements_this_rank = elements_this_rank, elements_per_rank = elements_per_rank, comm = comm, root = root)



def mpi_argsort(
    data: np.ndarray,
    use_index_dtype = np.int64,
    output_destination_ranks: bool = False,
    output_inplace_target_indexes: bool = False
) -> np.ndarray|tuple[np.ndarray, np.ndarray]|tuple[np.ndarray, np.ndarray, np.ndarray]:

    # Wait for all ranks to be ready
    mpi_barrier()

    # Compute the total length accross all ranks and index bounaries
    elements_per_rank: list[int] = MPI_Config.comm.allgather(data.shape[0])
    total_data_length = sum(elements_per_rank)
    rank_boundary_indexes: np.ndarray = np.array([0, *np.cumsum(elements_per_rank)], dtype = use_index_dtype) # boundarys are half-open [...)
    rank_global_start_index = rank_boundary_indexes[MPI_Config.rank]
    rank_global_end_index = rank_boundary_indexes[MPI_Config.rank + 1]

    # Create somwhere to store the final results of the argsort
    sorted_indexes = np.empty(shape = data.shape[0], dtype = use_index_dtype)
    if output_inplace_target_indexes:
        global_sorted_position_indexes = np.empty(shape = data.shape[0], dtype = use_index_dtype)
    if output_destination_ranks:
        destination_rank = np.empty(shape = data.shape[0], dtype = np.int64)

    # Sort the data locally
    # The smallest value accross all datasets must now be in the first position of the argsort of one of the ranks
    local_sort_indexes = np.argsort(data)
    local_sort_indexes_index: int|None = 0

    mpi_barrier()

    if MPI_Config.is_root:
        # List of rank indexes that can be filtered along with the selection options
        rank_indexes = np.arange(MPI_Config.comm_size)
        # Create a staging area to hold the smallest unselected value from each rank
        options = np.empty(shape = (MPI_Config.comm_size, *data.shape[1:]), dtype = data.dtype)
        # Create a mask that can be used to ignore ranks that have no remaining test values
        valid_options_mask = np.full(options.shape, True)

        # Set the first test value locally
        if data.shape[0] == 0:
            valid_options_mask[0] = False
        else:
            options[0] = data[local_sort_indexes[local_sort_indexes_index]]

        # Check all other ranks for data being avalible then recive the first value from such ranks
        for rank in range(0, MPI_Config.comm_size):
            mpi_barrier()
            # Ignore the root rank as that has already been handled
            if rank == MPI_Config.root:
                continue
            has_value_avalible = MPI_Config.comm.recv(source = rank)
            mpi_barrier()
            if has_value_avalible:
                options[rank] = MPI_Config.comm.recv(source = rank)
            else:
                valid_options_mask[rank] = False

    else:
        # Loop over all non-root ranks
        for rank in range(0, MPI_Config.comm_size):
            mpi_barrier()
            # Ignore the root rank as that has already been handled
            if rank == MPI_Config.root:
                continue
            # Wait for this rank's turn
            if MPI_Config.rank == rank:
                if data.shape[0] == 0:
                    # No data avalible
                    MPI_Config.comm.send(False, dest = MPI_Config.root)
                    mpi_barrier()
                else:
                    # Data is avalible so confirm this then send the smallest element
                    MPI_Config.comm.send(True, dest = MPI_Config.root)
                    mpi_barrier()
                    MPI_Config.comm.send(data[local_sort_indexes[local_sort_indexes_index]], dest = MPI_Config.root)
            else:
                # Everyone needs to call the barrier
                mpi_barrier()

    mpi_barrier()

    # Select the lowest element a number of times equal to the number of total data elements
    for sorted_element_index in range(total_data_length):

        # Select the lowest avalible option on the root rank then update all ranks with the rank that sent the smallest value
        rank_with_next_value: int
        if MPI_Config.is_root:
            rank_with_next_value = rank_indexes[valid_options_mask][np.argmin(options[valid_options_mask])]
        synchronyse("rank_with_next_value")

        global_index_of_selected_value: int

        # If this rank is the one with the current smallest value
        if MPI_Config.rank == rank_with_next_value:
            # Update the result for the current smallest item on this rank
            global_index_of_selected_value = rank_global_start_index + local_sort_indexes[local_sort_indexes_index]
            if output_inplace_target_indexes:
                global_sorted_position_indexes[local_sort_indexes[local_sort_indexes_index]] = sorted_element_index
            if output_destination_ranks:
                destination_rank[local_sort_indexes[local_sort_indexes_index]] = np.where(rank_boundary_indexes <= sorted_element_index)[0][-1]
            # Move to the next item
            local_sort_indexes_index += 1
            # If there is no next item, scrub the invalid index
            if local_sort_indexes_index >= data.shape[0]:
                local_sort_indexes_index = None

            # If the root rank has the selected value, just do all the updates locally
            if MPI_Config.is_root:
                if local_sort_indexes_index is not None:
                    options[0] = data[local_sort_indexes[local_sort_indexes_index]]
                else:
                    valid_options_mask[0] = False
                # Every other rank will have called an extra barrier to account for the usual status check call, so call anyway to match
                mpi_barrier()

            # If not the root rank, send the next lowest value to the root
            else:
                if local_sort_indexes_index is not None:
                    MPI_Config.comm.send(True, dest = MPI_Config.root)
                    mpi_barrier()
                    MPI_Config.comm.send(data[local_sort_indexes[local_sort_indexes_index]], dest = MPI_Config.root)
                else:
                    MPI_Config.comm.send(False, dest = MPI_Config.root)
                    mpi_barrier()

        # If the root rank does not contain the selected value, it still needs to recive an updated value
        elif MPI_Config.is_root:
            has_value_avalible = MPI_Config.comm.recv(source = rank_with_next_value)
            mpi_barrier()
            if has_value_avalible:
                options[rank_with_next_value] = MPI_Config.comm.recv(source = rank_with_next_value)
            else:
                valid_options_mask[rank_with_next_value] = False

        else:
            # Uninvolved ranks still need to call the barrier!
            mpi_barrier()

        mpi_barrier()

        synchronyse("global_index_of_selected_value", root = rank_with_next_value)

        mpi_barrier()

        if rank_global_start_index <= sorted_element_index and sorted_element_index < rank_global_end_index:
            sorted_indexes[sorted_element_index - rank_global_start_index] = global_index_of_selected_value

        mpi_barrier()

    mpi_barrier()

    result = [sorted_indexes]
    if output_destination_ranks:
        result.append(destination_rank)
    if output_inplace_target_indexes:
        result.append(global_sorted_position_indexes)
    return result[0] if len(result) == 1 else tuple(result)



def mpi_sort(
        data: np.ndarray,
        order: tuple[np.ndarray, np.ndarray]|None = None,
        result_data_buffer: np.ndarray|None = None,
        assume_memory_limitation: bool = False
) -> np.ndarray:
    """
    Sort a set of data in parallel.

    Parameters:

                           `numpy.ndarray` `data`                     -> The data to be sorted (along axis = 0)

        (`numpy.ndarray`, `numpy.ndarray`) `order`                    -> The ordering information from `mpi_argsort(...)[1:]` (target_ranks, inplace_sort_indexes) (optional)

                           `numpy.ndarray` `result_data_buffer`       -> Buffer to place results into (optional)

                                    `bool` `assume_memory_limitation` -> Perform the sort in a memory-efficient way, at the cost of speed (defaults to False - fastest)

    Returns:
        `numpy.ndarray` -> The data in a sorted order distributed accross each rank using the same distribution as the input data
    """
    if order is None:
        order = mpi_argsort(data, output_destination_ranks = True, output_inplace_target_indexes = True)[1:]

    sorted_data: np.ndarray
    unscrambling_indexes: np.ndarray
    if result_data_buffer is not None:
        sorted_data = result_data_buffer
    if not assume_memory_limitation:
        # Might as well request the memory in parallel
        if result_data_buffer is None:
            sorted_data = np.empty_like(data)
        unscrambling_indexes = np.empty(shape = data.shape[0], dtype = np.int64)

    for target_rank in range(MPI_Config.comm_size):
        mpi_barrier()

        rank_is_target = MPI_Config.check_is_root(target_rank)
        send_mask = order[0] == target_rank

        if assume_memory_limitation and rank_is_target:
            # Avoid requesting the memory until absolutely nessessary
            if result_data_buffer is None:
                sorted_data = np.empty_like(data)
            unscrambling_indexes = np.empty(shape = data.shape[0], dtype = np.int64)

        mpi_gather_array(    data[send_mask], root = target_rank, target_buffer =          sorted_data if rank_is_target else None)
        mpi_gather_array(order[1][send_mask], root = target_rank, target_buffer = unscrambling_indexes if rank_is_target else None)

        if assume_memory_limitation:
            # Do the unscramble now to avoid unnessessary memory being held
            # This will be slower as all ranks must wait for this to be completed
            if rank_is_target:
                unscrambled = np.empty_like(sorted_data)
                unscrambled[unscrambling_indexes - unscrambling_indexes.min()] = sorted_data
                sorted_data = unscrambled[:]
                del unscrambling_indexes
                del unscrambled

    mpi_barrier()

    if not assume_memory_limitation:
        # Do the unscramble at the end
        # This will be faster as this can be done in parallel, but uses more memory as the unscramble indexes are stored accross all calls
        unscrambled = np.empty_like(sorted_data)
        unscrambled[unscrambling_indexes - unscrambling_indexes.min()] = sorted_data
        sorted_data = unscrambled[:]
        del unscrambling_indexes
        del unscrambled

    return sorted_data



def mpi_calculate_reorder(
    input_ids: np.ndarray,
    output_ids: np.ndarray,
    assume_memory_limitation: bool = False
) -> tuple[np.ndarray, np.ndarray]:
    order = mpi_argsort(input_ids, output_destination_ranks = True, output_inplace_target_indexes = True)
    reverse_order = mpi_argsort(order[0], output_destination_ranks = True, output_inplace_target_indexes = True)
    del order

    target_sort_order = mpi_argsort(output_ids, output_destination_ranks = True, output_inplace_target_indexes = True)
    reverse_target_sort_order = mpi_argsort(target_sort_order[0], output_destination_ranks = True, output_inplace_target_indexes = True)
    del target_sort_order

    return (
        mpi_sort(data = reverse_target_sort_order[1], order = reverse_order[1:], assume_memory_limitation = assume_memory_limitation),
        mpi_sort(data = reverse_target_sort_order[2], order = reverse_order[1:], assume_memory_limitation = assume_memory_limitation)
    )



def mpi_reorder(
    input_ids: np.ndarray,
    input_data: np.ndarray,
    output_ids: np.ndarray,
    output_data_buffer: np.ndarray|None = None,
    assume_memory_limitation: bool = False
) -> np.ndarray:
    order = mpi_calculate_reorder(input_ids, output_ids)
    return mpi_sort(
        data = input_data,
        order = order,
        result_data_buffer = output_data_buffer,
        assume_memory_limitation = assume_memory_limitation
    )



def mpi_apply_reorder(
    input_data: np.ndarray,
    order: tuple[np.ndarray, np.ndarray],
    output_data_buffer: np.ndarray|None = None,
    assume_memory_limitation: bool = False
) -> np.ndarray:
    """
    Wrapper for `mpi_sort` using a subset of the parameter names from `mpi_reorder`.

    `input_data` -> `data`

    `order` -> `order`

    `output_data_buffer` -> `result_data_buffer`

    `assume_memory_limitation` -> `assume_memory_limitation`
    """
    return mpi_sort(
        data = input_data,
        order = order,
        result_data_buffer = output_data_buffer,
        assume_memory_limitation = assume_memory_limitation
    )




from ._IPlayingCard import IPlayingCard
from ._PlayingCard import PlayingCard

from ._CardGroup import CardGroup
from ._Deck import Deck
from ._CardStack import CardStack
from ._CardQueue import CardQueue



import re

# No "isBool" exists as any object can be converted to a boolean

def isChar(testCase):
    return isString(testCase) and len(str(testCase)) == 1

def isString(testCase):
    try:
        str(testCase)
        return True
    except:
        return False

def isAlpha(testCase):
    return isString(testCase) and re.match("^[a-zA-Z_']+$", testCase) is not None

def isInt(testCase):
    try:
        int(testCase)
    except:
        return False

    return re.match("^[0-9\-]+$", str(testCase)) is not None

def isFloat(testCase):
    try:
        float(testCase)
    except:
        return False

    return re.match("^[0-9\.\-]+$", str(testCase)) is not None

def inRange(testCase, lower, upper, inside = True):
    if upper == lower:
        raise ValueError("The arguments for paramiters min and max are the same. They must be different as min is inclusive and max is exclusive.")

    return (lower <= testCase) and (testCase < upper) == inside

def isEven(testCase):
    return float(testCase // 2) == float(testCase / 2)

def isOdd(testCase):
    return not isEven(testCase)

def isFactor(testCase, value):
    return isInt(testCase) and isInt(value) and testCase != 0 and isInt(value / testCase)

def isMultiple(testCase, value):
    return isInt(testCase) and isInt(value) and value != 0 and isInt(testCase / value)

def isEmail(testCase):
    return isString(testCase) and re.match("^(mailto:)?[a-zA-Z0-9]+(\.[a-zA-Z0-9]+)?@[a-zA-Z]+(\.[a-zA-Z]+)+$", str(testCase)) is not None

def isURL(testCase):#TODO: some avalable characters still missing
    return isURI(testCase) and re.match("^https?:", str(testCase)) is not None

def isURI(testCase):
    #                                        |scheme              |authority                                                                                                              |path                                               |query                                                               |fragment
    return isString(testCase) and re.match("^[a-z][a-z1-9\+\.\-]*:(//([a-zA-Z1-9%]+(\.[a-zA-Z1-9%]+)*(:[a-zA-Z1-9%]+)?@)?(([a-zA-Z1-9%]+(\.[a-zA-Z1-9%]+)+)|(\[[a-z1-9]{4}(:[a-z1-9]{4}){7}\]))?(:[0-9]+)?)?(/[a-zA-Z0-9%]+(/[a-zA-Z0-9%]+)+(/|\.[a-zA-Z0-9%]+)?)(\?[a-zA-Z0-9%]+=[a-zA-Z0-9%]+((\&|;)[a-zA-Z0-9%]+=[a-zA-Z0-9%]+)*)?(#[a-zA-Z1-9%]+)?$", str(testCase)) is not None
    #                                                                |userinfo                         |host                                                                  |port

def isURIScheme(testCase):
    isString(testCase) and re.match("^[a-z][a-z1-9\+\.\-]*:", str(testCase)) is not None

def isURIAuthority(testCase):
    raise NotImplementedError("This method has not yet been implemented.")

def isURIPath(testCase):
    raise NotImplementedError("This method has not yet been implemented.")

def isURIQuery(testCase):
    raise NotImplementedError("This method has not yet been implemented.")

def isURIFragment(testCase):
    raise NotImplementedError("This method has not yet been implemented.")


class ISpinner(object):
    """
    Interface for all Spinner objects.
    """

    @property
    def Labels(self) -> list:
        """
        Array of side labels
        """
        raise NotImplementedError()

    @property
    def Sides(self) -> int:
        """
        Number of sides/labels
        """
        raise NotImplementedError()

    def ContextSpin(self):
        """
        Randomly selects a side and retuns the side and the corisponding label.
        """
        raise NotImplementedError()

    def Spin(self):
        """
        Randomly selects a side and retuns the corisponding label.
        """
        raise NotImplementedError()

    def IndexSpin(self):
        """
        Randomly selects a side and retuns the side number.
        """
        raise NotImplementedError()


import sys
import traceback
from typing import TypeVar, Generic, Any, Union, Collection, Sequence, List, Tuple, Dict, Callable
from abc import ABC, abstractmethod

from ..IO.Text._console import Console# print_info, print_verbose_info, print_warning, print_verbose_warning, print_error, print_verbose_error, print_debug
from .._global_settings import settings_object as _settings_object
from ..MPI import get_mpi_rank
from ..Tools._async import start_main_async
from ..IO.Text._stopwatch import Stopwatch

T = TypeVar("T")
U = TypeVar("U")

from types import ModuleType
from ..Data import AuthorInfomation as _AuthorInfomation, VersionInfomation as _VersionInfomation
from datetime import date as _date
from functools import wraps



def ScriptWrapper_passthrough_converter(value: str) -> str:
    return value

def ScriptWrapper_bool_converter(value: str) -> bool:
    return value.lower() in ("true", "t", "yes", "y", "1", "accept", "allow", "confirm")

def ScriptWrapper_filter_converter(value: str) -> Tuple[bool, ...]:
    return tuple(ScriptWrapper.bool_converter(element) for element in value.lower())

#def ScriptWrapper_reverse_flag_converter(value: bool) -> bool:
#    return not value

def ScriptWrapper_make_list_converter(seperator = ",", item_converters: Union[Sequence[Callable[[str], Any]], Callable[[str], Any], None] = None):
    if item_converters is None:
        item_converters = ScriptWrapper.passthrough_converter
    def list_converter(value: str) -> Any:
        if not isinstance(item_converters, Sequence):
            return [item_converters(item) if item != "" or item_converters == ScriptWrapper.passthrough_converter or item_converters == str else None for item in value.split(seperator)]
        else:
            converters_for_use: List[Callable[[str], Any]] = list(item_converters)
            strings = value.split(seperator)
            while len(converters_for_use) < len(strings):
                converters_for_use.append(ScriptWrapper.passthrough_converter)
            return [converters_for_use[i](strings[i]) if strings[i] != "" or converters_for_use[i] == ScriptWrapper.passthrough_converter or converters_for_use == str else None for i in range(len(strings))]
    return list_converter



class ScriptWrapper_ParamSpec(List["ScriptWrapper_ParamBase"]):
    """
    The parameters for a ScriptWrapper instance.
    """

    def __init__(
            self,
            *params: "ScriptWrapper_ParamBase",
            common_params_have_short_names: bool = True,
            custom_help_param: Union["ScriptWrapper_Flag", None] = None,
            custom_debug_param: Union["ScriptWrapper_Flag", None] = None,
            custom_verbose_param: Union["ScriptWrapper_Flag", None] = None
        ) -> None:
        super().__init__()
        self.__help_param    = custom_help_param if custom_help_param is not None else make_help_flag(long_name_only = not common_params_have_short_names)
        self.__debug_param   = custom_debug_param if custom_debug_param is not None else make_debug_flag(long_name_only = not common_params_have_short_names)
        self.__verbose_param = custom_verbose_param if custom_verbose_param is not None else make_verbose_flag(long_name_only = not common_params_have_short_names)
        self.__special_param_names = (self.__help_param.name, self.__debug_param.name, self.__verbose_param.name)
        self.__special_param_short_names = (self.__help_param.short_name, self.__debug_param.short_name, self.__verbose_param.short_name)
        self.add(self.__help_param   )
        self.add(self.__debug_param  )
        self.add(self.__verbose_param)
        self.add(*params)

    @property
    def help_param(self) -> "ScriptWrapper_Flag":
        return self.__help_param
    @property
    def debug_param(self) -> "ScriptWrapper_Flag":
        return self.__debug_param
    @property
    def verbose_param(self) -> "ScriptWrapper_Flag":
        return self.__verbose_param
    @property
    def special_param_names(self) -> Tuple[str, str, str]:
        return self.__special_param_names
    def special_param_short_names(self) -> Tuple[Union[str, None], Union[str, None], Union[str, None]]:
        return self.__special_param_short_names

    def add(self, *params: "ScriptWrapper_ParamBase"):
        """
        Insert parameter definitions.

        ScriptWrapper_ParamSpec_Error will be raised if an invalid addition is requested.
        In the event of a ScriptWrapper_ParamSpec_Error, no additions will be performed.
        """
        
        try:
            for param in params:
                for test_param in self:
                    if len(self) > 0 and isinstance(param, ScriptWrapper_PositionalParam) and self[-1].name not in self.__special_param_names and not isinstance(self[-1], ScriptWrapper_PositionalParam):
                        raise ScriptWrapper_ParamSpec_Error(f"Attempted to register a positional parameter ({param.name}) after registering a number of non-positional parameters.\nAll positional parameters must be registered in-order and before any non-positional parameters.")
                    if param.name == test_param.name:
                        raise ScriptWrapper_ParamSpec_Error(f"Attempted to register two parameters with the name \"{param.name}\".")
                    if param.short_name == test_param.short_name and param.short_name is not None:
                        raise ScriptWrapper_ParamSpec_Error(f"Attempted to register two parameters with the short name \"{param.short_name}\".")
                    if param.target_parameter_name == test_param.target_parameter_name:
                        if not param.can_conflict_with(test_param):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} that targets the same function parameter ({param.target_parameter_name}) as {test_param.name}, but does not indicate the later as a conflict.\nParameters that target the same function parameter must indicate each other as conflicts!")
                        elif not test_param.can_conflict_with(param):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} that targets the same function parameter ({param.target_parameter_name}) as {test_param.name}, but the latter does not indicate the former as a conflict.\nParameters that target the same function parameter must indicate each other as conflicts!")
                        if isinstance(param, ScriptWrapper_RequiredParam):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} that targets the same function parameter ({param.target_parameter_name}) as {test_param.name}, but the former is a required parameter.\nRequired parameters must be the sole parameter definition for a particular target function parameter.")
                        elif isinstance(test_param, ScriptWrapper_RequiredParam):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} that targets the same function parameter ({param.target_parameter_name}) as {test_param.name}, but the latter is a required parameter.\nRequired parameters must be the sole parameter definition for a particular target function parameter.")
                        if param.requires(test_param):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} that targets the same function parameter ({param.target_parameter_name}) as {test_param.name}, but the former requires the latter.\nParameter definitions that target the same function parameter cannot also require each other!")
                        elif param.is_required_by(test_param):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} that targets the same function parameter ({param.target_parameter_name}) as {test_param.name}, but the latter requires the former.\nParameter definitions that target the same function parameter cannot also require each other!")
                        if param.default is not None and test_param.default is not None:
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} that targets the same function parameter ({param.target_parameter_name}) as {test_param.name}, but both specify a non-null default value.\nParameter definitions that target the same function parameter cannot include more than one non-null default value!")
                    if param.can_conflict_with(test_param):
                        if isinstance(test_param, ScriptWrapper_RequiredParam):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} with a required parameter ({test_param.name}) as a conflict.\nConflicting parameters cannot be required parameters!")
                        elif param.is_required_by(test_param):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} with {test_param.name} as a conflict, however {test_param.name} lists {param.name} as a requirement.\nConflicting parameters cannot also require each other!")
                    if test_param.can_conflict_with(param):
                        if isinstance(param, ScriptWrapper_RequiredParam):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} (a required parameter) but ({test_param.name}) indicates {param.name} is a conflict.\nConflicting parameters cannot be required parameters!")
                        elif param.requires(test_param):
                            raise ScriptWrapper_ParamSpec_Error(f"Attempted to register parameter {param.name} with {test_param.name} as a requirement, however {test_param.name} lists {param.name} as a conflict.\nConflicting parameters cannot also require each other!")
                param._regester_spec(self)
                self.append(param)
        except Exception as e:
            for param in params:
                param._regester_spec(None)
                try:
                    self.remove(param)
                except: pass
            raise e
        
    def get_keyword_arguments(self) -> Dict[str, Union[Any, None]]:
        kwargs: Dict[str, Union[Any, None]] = {}
        for param in self:
            if param.name in self.special_param_names:
                continue # Special params shouls not be passed to the target function.
            if param.target_parameter_name not in kwargs:
                kwargs[param.target_parameter_name] = None
            if param.has_non_null_value:
                kwargs[param.target_parameter_name] = param.value
        return kwargs
        
    def reset(self) -> None:
        """
        Reset all parameters back to their starting values.
        """
        
        for p in self:
            p._reset()
    
    #TODO: handle __add__ and other list type features!

class ScriptWrapper_ParamSpec_Error(Exception):
    """
    An error arrising from a ScriptWrapper_ParamSpec object.
    """

class ScriptWrapper_ParamBase(ABC, Generic[T]):
    """
    
    """

    def __init__(
                    self,
                    name:                str,
                    short_name:          Union[str, None]                                               = None,
                    sets_param:          Union[str, None]                                               = None,
                    default_value:       Union[T, None]                                                 = None,
                    conversion_function: Union[Callable[[str], T], None]                                = None,
                    conflicts:           Union[Collection[Union[str, "ScriptWrapper_ParamBase"]], None] = None,
                    requirements:        Union[Collection[Union[str, "ScriptWrapper_ParamBase"]], None] = None,
                    description:         Union[str, None]                                               = None
                ) -> None:
        self.__name: str = name
        self.__short_name: Union[str, None] = short_name
        self.__parameter_name_to_set: str = sets_param if sets_param is not None else name.replace("-", "_")
        self.__default_value: Union[T, None] = default_value
#        self.__conversion_function: Callable[[str], T] = conversion_function if conversion_function is not None else ScriptWrapper.passthrough_converter
        self.__conversion_function: Callable[[str], T] = conversion_function if conversion_function is not None else ScriptWrapper_passthrough_converter
        self.__conflicting_param_names: Tuple[str, ...] = tuple((v.name if isinstance(v, ScriptWrapper_ParamBase) else v) for v in conflicts   ) if conflicts is not None else tuple()
        self.__requirement_param_names: Tuple[str, ...] = tuple((v.name if isinstance(v, ScriptWrapper_ParamBase) else v) for v in requirements) if requirements is not None else tuple()
        self.__description: str = description if description is not None else ""

        self.__value: Union[T, None] = self.__default_value
        self.__value_explicitly_set = False

        self.__spec: Union[ScriptWrapper_ParamSpec, None] = None

        invalid_params = set(self.__conflicting_param_names).intersection(set(self.__requirement_param_names))
        if len(invalid_params) > 0:
            raise ValueError(f"Conflicts and requirements contained common parameters: {invalid_params}")

    @property
    def name(self) -> str:
        """
        Name of the parameter.

        This parameter may be specified by preceding the name with two dashes (e.g. '--<param name> <appropriate value>').
        """
        return self.__name

    @property
    def short_name(self) -> Union[str, None]:
        """
        Short string alternate name for the parameter.

        This parameter may be specified by preceding the name with a single dash (e.g. '-<short param name> <appropriate value>').
        """
        return self.__short_name

    @property
    def target_parameter_name(self) -> str:
        """
        Name of the function kwarg set by this parameter.
        """
        return self.__parameter_name_to_set

    @property
    def default(self) -> Union[T, None]:
        """
        Default value if this parameter is not specified.
        """
        return self.__default_value

    @property
    def conversion_function(self) -> Callable[[str], T]:
        """
        Function to be used for converting inputs to the intended type.
        """
        return self.__conversion_function

    @property
    def conflicts(self) -> Tuple[str, ...]:
        """
        Parameters that cannot be set if this parameter is used.
        """
        return self.__conflicting_param_names
    
    @property
    def requirements(self) -> Tuple[str, ...]:
        """
        Parameters that must be set if this parameter is used.
        """
        return self.__requirement_param_names

    @property
    def description(self) -> str:
        """
        Description of the parameter.
        """
        return self.__description

    @property
    def value(self) -> Union[T, None]:
        """
        Current assigned value.
        """
        return self.__value

    def __str__(self) -> str:
        return f"ScriptWrapper Parameter: {self.__name}\n{self.__description}"
    
    def _regester_spec(self, param_spec: Union[ScriptWrapper_ParamSpec, None]) -> None:
        self.__spec = param_spec

    @property
    def _spec(self) -> Union[ScriptWrapper_ParamSpec, None]:
        return self.__spec
    
    def can_conflict_with(self, test_param: "ScriptWrapper_ParamBase") -> bool:
        """
        Can this parameter conflict with another?
        """
        return test_param.name in self.__conflicting_param_names

    def conflicts_with(self, test_param: "ScriptWrapper_ParamBase") -> bool:
        """
        Does this parameter conflict with another, given their current state?
        """
        return self.can_conflict_with(test_param) and self.is_explicitly_set and test_param.is_explicitly_set
    
    def requires(self, test_param: "ScriptWrapper_ParamBase") -> bool:
        """
        Does this parameter require another?
        """
        return test_param.name in self.__requirement_param_names
    
    def is_required_by(self, test_param: "ScriptWrapper_ParamBase") -> bool:
        """
        Does another parameter require this one?
        """
        return test_param.requires(self)
    
    @property
    def has_non_null_value(self) -> bool:
        """
        Does the param have a value that isn't None.
        """
        return self.__value is not None
    
    @property
    def is_explicitly_set(self) -> bool:
        """
        Has a value been assigned to the param.
        """
        return self.__value_explicitly_set
    
    def _set(self, value: T, /) -> None:
        self.__value = value
        self.__value_explicitly_set = True

    def _reset(self):
        self.__value = self.__default_value
        self.__value_explicitly_set = False

    @abstractmethod
    def set_from_args(self, named_arguments: Dict[str, str]) -> bool:
        """
        Attempt to set the value of this parameter given the avalible argument strings.
        """
        raise NotImplementedError("Attempted to call an abstract method.")
    
    def validate(self) -> None:
        """
        Ensure the value assigned to this patameter is valid after parsing all parameters.
        """
        if self.__spec is None:
            raise ScriptWrapper_Param_ValidityError(self, f"Parameter {self.name} was not assigned to a ScriptWrapper_ParamSpec object.")
        if self.is_explicitly_set:
            for test_param in self.__spec:
                if self.conflicts_with(test_param):
                    raise ScriptWrapper_Param_ConflictError(self, test_param)
                elif self.requires(test_param) and not test_param.is_explicitly_set:
                    raise ScriptWrapper_Param_MissingRequirementError(self, test_param)
    
class ScriptWrapper_Param_ValidityError(Exception):
    """
    Parameter definition was invalid after parsing of arguments.
    """
    def __init__(self, param_object: ScriptWrapper_ParamBase, message: str):
        self.raising_parameter: ScriptWrapper_ParamBase = param_object
        super().__init__(message)
class ScriptWrapper_Param_ValueError(ScriptWrapper_Param_ValidityError, ValueError):
    """
    Parameter definition was assigned an invalid value.
    """
    def __init__(self, param: ScriptWrapper_ParamBase, value_string: str):
        super().__init__(param, f"Parameter {param.name} was passed an invalid value: \"{value_string}\"")
class ScriptWrapper_Param_TypeError(ScriptWrapper_Param_ValidityError, TypeError):
    """
    Parameter definition parsed an argument and produced an invalid type.
    """
    def __init__(self, param: ScriptWrapper_ParamBase):
        super().__init__(param, f"Parameter {param.name} produced a value with an invalid type: {type(param.value)}")
class ScriptWrapper_Param_UnsetError(ScriptWrapper_Param_ValidityError):
    """
    Parameter definition required a value, but none was passed as an argument.
    """
    def __init__(self, param: ScriptWrapper_ParamBase):
        super().__init__(param, f"Parameter {param.name} requires a value but none was provided.")
class ScriptWrapper_Param_ConflictError(ScriptWrapper_Param_ValidityError):
    """
    Parameter definition was assigned a value, along woth a conflicting parameter.
    """
    def __init__(self, param: ScriptWrapper_ParamBase, conflicting_param: ScriptWrapper_ParamBase):
        self.conflicting_parameter: ScriptWrapper_ParamBase = conflicting_param
        super().__init__(param, f"Parameter {param.name} conflicts with {conflicting_param.name} and both were set.")
class ScriptWrapper_Param_MissingRequirementError(ScriptWrapper_Param_ValidityError):
    """
    Parameter definition was assigned a value, however a required parameter was not.
    """
    def __init__(self, param: ScriptWrapper_ParamBase, required_param: ScriptWrapper_ParamBase):
        self.required_parameter: ScriptWrapper_ParamBase = required_param
        super().__init__(param, f"Parameter {param.name} requires {required_param.name}, however the later was not explicitly specified.")

class ScriptWrapper_ArgParamBase(ScriptWrapper_ParamBase[T]):
    def __init__(
                    self,
                    name:                str,
                    short_name:          Union[str, None]                                               = None,
                    sets_param:          Union[str, None]                                               = None,
                    default_value:       Union[Any, None]                                               = None,
                    conversion_function: Union[Callable[[str], T], None]                                = None,
                    type_check:          Union[type, None]                                              = None,
                    conflicts:           Union[Collection[Union[str, ScriptWrapper_ParamBase]], None]   = None,
                    requirements:        Union[Collection[Union[str, ScriptWrapper_ParamBase]], None]   = None,
                    description:         Union[str, None]                                               = None
                ) -> None:
        super().__init__(
                            name = name,
                      short_name = short_name,
                      sets_param = sets_param,
                   default_value = default_value,
             conversion_function = conversion_function,
                       conflicts = conflicts,
                    requirements = requirements,
                     description = description
        )
        self.__required_type: Union[type, None] = type_check
        if default_value is not None:
            self._set(default_value)

    @property
    def _type_to_check_against(self) -> Union[type, None]:
        """
        Type of which a valid resulting value must be an instance.
        """
        return self.__required_type

    def _set_from_raw(self, input_string: str, /) -> None:
        self._set(self.conversion_function(input_string))

    def set_from_args(self, named_arguments: Dict[str, str]) -> bool:
        if self.name in named_arguments:
            self._set_from_raw(named_arguments[self.name])
            return True
        return False

    def validate(self):
        if self._type_to_check_against is not None and self.is_explicitly_set and not isinstance(self.value, self._type_to_check_against):
            raise ScriptWrapper_Param_TypeError(self)
        super().validate()

class ScriptWrapper_OptionalParam(ScriptWrapper_ArgParamBase[T]):
    """
    
    """

class ScriptWrapper_RequiredParam(ScriptWrapper_ArgParamBase[T]):
    """
    
    """

    def __init__(
                    self,
                    name:                   str,
                    short_name:             Union[str, None]                = None,
                    sets_param:             Union[str, None]                = None,
                    conversion_function:    Union[Callable[[str], T], None] = None,
                    type_check:             Union[type, None]               = None,
                    description:            Union[str, None]                = None
                ) -> None:
        super().__init__(
                            name = name,
                      short_name = short_name,
                      sets_param = sets_param,
                   default_value = None,
             conversion_function = conversion_function,
                      type_check = type_check,
                       conflicts = None,
                    requirements = None,
                     description = description
        )

    def validate(self):
        if not self.is_explicitly_set:
            raise ScriptWrapper_Param_UnsetError(self)
        super().validate()

class ScriptWrapper_PositionalParam(ScriptWrapper_RequiredParam[T]):
    """
    A type of ScriptWrapper parameter that must be specified at in the parameter list before any optional parameters or flags.
    This type of parameter is not required to be specified by name (though may be if desired).
    These parameters have no default value, and must be assigned a value.
    """

    def set_from_positional_arg(self, arg_string: str) -> None:
        self._set_from_raw(arg_string)

class ScriptWrapper_Flag(ScriptWrapper_ParamBase[bool]):
    """
    A type of ScriptWrapper parameter that takes no argument and toggles a single boolean state.
    """

    def __init__(
                    self,
                    name:           str,
                    short_name:     Union[str, None]                                                = None,
                    sets_param:     Union[str, None]                                                = None,
                    inverted:       bool                                                            = False,
                    conflicts:      Union[Collection[Union[str, ScriptWrapper_ParamBase]], None]    = None,
                    requirements:   Union[Collection[Union[str, ScriptWrapper_ParamBase]], None]    = None,
                    description:    Union[str, None]                                                = None
                ) -> None:
        super().__init__(
                            name = name,
                      short_name = short_name,
                      sets_param = sets_param,
                   default_value = bool(inverted), # Cast to a bool just in case attempt is made to set field with an integer
             conversion_function = None,
                       conflicts = conflicts,
                    requirements = requirements,
                     description = description
        )
        self.__inverted = inverted

    @property
    def inverted(self) -> bool:
        """
        Type of which a valid resulting value must be an instance.
        """
        return self.__inverted

    def set_from_args(self, named_arguments: Dict[str, str]) -> bool:
        if self.name in named_arguments:
            self._set(not self.__inverted)
            return True
        return False



def make_help_flag(name = "help", short_name = "h", long_name_only = False):
    return ScriptWrapper_Flag(
               name = name,
         short_name = None if long_name_only else short_name,
         sets_param = None,
           inverted = False,
        description = "Display this docstring."
    )
def make_debug_flag(name = "debug", short_name = "d", long_name_only = False):
    return ScriptWrapper_Flag(
               name = name,
         short_name = None if long_name_only else short_name,
         sets_param = None,
           inverted = False,
        description = "Display debug infomation."
    )
def make_verbose_flag(name = "verbose", short_name = "v", long_name_only = False):
    return ScriptWrapper_Flag(
               name = name,
         short_name = None if long_name_only else short_name,
         sets_param = None,
           inverted = False,
        description = "Display progression infomation."
    )



class ScriptWrapper(object):
    """
    Singelton that handles command argument parsing for the currently running program.
    Constructor Arguments:

               (ScriptWrapper.ParamSpec | list | None) parameters               --> Definitions for parsing arguments to the script.
                                                                                        Use: ScriptWrapper.OptionalParam
                                                                                             ScriptWrapper.RequiredParam
                                                                                             ScriptWrapper.PositionalParam
                                                                                             ScriptWrapper.Flag

                                          (str | None) command                  --> Command used to execute script.

                                          (str | None) filename                 --> Name of the script file.

               (ScriptWrapper.AuthorInfomation | None) authors                  --> Information about the authors.

        (ScriptWrapper.VersionInfomation | str | None) version                  --> Script version.
        
                                (datetime.date | None) edit_date                --> Date of last edit.

                                          (str | None) description              --> Information about what the script does and how to use it.

                 (List[types.ModuleType | str] | None) dependancies             --> Libraries needed to run the script.

                                    (List[str] | None) usage_paramiter_examples --> Example argument configurations.

    Public Methods:
               parse_arguments
               run
               run_with_async
        static passthrough_converter
        static bool_converter
        static filter_converter
        static make_list_converter
#    """
    
    # Convinience type aliases for accessing dependant types
    AuthorInfomation  = _AuthorInfomation
    VersionInfomation = _VersionInfomation
    ParamSpec         = ScriptWrapper_ParamSpec
    OptionalParam     = ScriptWrapper_OptionalParam
    RequiredParam     = ScriptWrapper_RequiredParam
    PositionalParam   = ScriptWrapper_PositionalParam
    Flag              = ScriptWrapper_Flag

    #TODO: set this up to enforce a singleton
    #__initialised = False
    #__instance = None
    #
    #def __new__(cls, *args, **kwargs):
    #    if ScriptWrapper.__initialised:
    #        #raise RuntimeError("A running program may only define one ScriptWrapper instance.")
    #        return ScriptWrapper.__instance
    #    else:
    #        ScriptWrapper.__initialised = True
    #        ScriptWrapper.__instance = super(ScriptWrapper, cls).__new__(cls)
    #        return ScriptWrapper.__instance

    def __init__(
                    self,
                    parameters:               Union[ScriptWrapper_ParamSpec,          List[ScriptWrapper_ParamBase], None] = None,
                    command:                  Union[str,                                                             None] = None,
                    filename:                 Union[str,                                                             None] = None,
                    authors:                  Union[Sequence[_AuthorInfomation],                                     None] = None,
                    version:                  Union[_VersionInfomation,               str,                           None] = None,
                    edit_date:                Union[_date,                                                           None] = None,
                    description:              Union[str,                                                             None] = None,
                    dependancies:             Union[Sequence[Union[ModuleType, str]],                                None] = None,
                    usage_paramiter_examples: Union[Sequence[str],                                                   None] = None
                ) -> None:
        
        if usage_paramiter_examples is not None and len(list(v for v in usage_paramiter_examples)) > 0 and command is None and filename is None:
            raise ValueError("Unable to render usage examples without either a command name or file name.")
        
        if parameters is None:
            parameters = ScriptWrapper_ParamSpec()
        elif not isinstance(parameters, ScriptWrapper_ParamSpec):
            parameters = ScriptWrapper_ParamSpec(*parameters)

        if version is None:
            version = _VersionInfomation.from_string("0.0.0")
        elif isinstance(version, str):
            version = _VersionInfomation.from_string(version)

        self.__stopwatch: Stopwatch = Stopwatch(command if command is not None else filename if filename is not None else "Script")

        self.__parameters:               ScriptWrapper_ParamSpec       = parameters
        self.__command_name:             Union[str, None]              = command
        self.__file_name:                Union[str, None]              = filename
        self.__author_info:              Tuple[_AuthorInfomation, ...] = tuple(authors) if authors is not None else tuple()
        self.__version:                  _VersionInfomation            = version
        self.__edit_date:                Union[_date, None]            = edit_date
        self.__description:              Union[str, None]              = description
        self.__import_dependancies:      Tuple[str, ...]               = tuple((dep.__name__ if isinstance(dep, ModuleType) else dep) for dep in dependancies) if dependancies is not None else tuple()
        self.__usage_paramiter_examples: Tuple[str, ...]               = tuple(usage_paramiter_examples) if usage_paramiter_examples is not None else tuple()

        self.__help_string: str = ""
        self.__render_help_string()

        self.__parameter_lookup_by_name: Dict[str, ScriptWrapper_ParamBase] = { p.name : p for p in self.__parameters }
        self.__parameter_lookup_by_short_name: Dict[str, ScriptWrapper_ParamBase] = { p.short_name : p for p in self.__parameters if p.short_name is not None }
        self.__arguments_have_been_parsed: bool = False
        self.__raw_arguments: Union[Sequence[str], None] = None

        self.use_MPI = False
        self.MPI_print_non_root_process_errors = False

    def __render_help_string(self):
        self.__help_string = ""
        if self.__command_name is not None:
            self.__help_string     += f"\nCommand: {self.__command_name}"
        if self.__file_name is not None:
            self.__help_string     += f"\nFile:    {self.__file_name}"
        self.__help_string         += f"\nVersion: {self.__version.to_string(minimum_length = True)}"
        if self.__edit_date is not None:
            self.__help_string     += f"\nDate:     {self.__edit_date.strftime(_settings_object.date_format)}"
        if len(self.__author_info) > 0:
            self.__help_string     +=  "\nAuthor:  " if len(self.__author_info) == 1 else "\nAuthors: "
            for i, author in enumerate(self.__author_info):
                self.__help_string +=  ("\n         " if i > 0 else "") + f"{author.given_name} {author.family_name}" + ((" (" + (author.email_address if author.email_address is not None else "") + (" | " if author.email_address is not None and author.website_url is not None else "") + (author.website_url if author.email_address is not None else "") + ")") if author.email_address is not None or author.website_url is not None else "")
        if self.__description is not None:
            self.__help_string     += f"\n\n{self.__description}"
        self.__help_string += """

Commandline arguments & flags ( p = required positional parameter,
                                r = required parameter,
                               f  = optional flag,
                               f+ = optional parameter with an argument):"""
        name_max_width = max([len(parameter_definition.name) for parameter_definition in self.__parameters])
        short_name_max_width = max([len(parameter_definition.short_name) if parameter_definition.short_name is not None else 0 for parameter_definition in self.__parameters])
        indent_spacing = name_max_width + ((short_name_max_width + 5) if short_name_max_width > 0 else 0) + 20
        for parameter_definition in self.__parameters:
            self.__help_string     +=  ("\n\n    ({}) --{}{}" + (" || " if short_name_max_width > 0 else "") + "{}{}{}").format(
                (" p" if isinstance(parameter_definition, self.PositionalParam) else (" r" if isinstance(parameter_definition, self.RequiredParam) else ("f " if isinstance(parameter_definition, self.Flag) else "f+"))),
                parameter_definition.name,
                (name_max_width - len(parameter_definition.name)) * " ",
                "" if parameter_definition.short_name is None else f"-{parameter_definition.short_name}",
                " " * (short_name_max_width + (1 if short_name_max_width > 0 else 0) - ((len(parameter_definition.short_name) + 1) if parameter_definition.short_name is not None else 0)),
                (" --> " + parameter_definition.description.replace("\n", "\n" + (" " * indent_spacing))) if parameter_definition.description is not None else ""
            )
        if len(self.__import_dependancies) > 0:
            self.__help_string     +=  "\n\nDependancies:"
            for dependancy in self.__import_dependancies:
                self.__help_string += f"\n    {dependancy}"
        if len(self.__usage_paramiter_examples) > 0:
            self.__help_string     +=  "\n\nExample Usage:"
            for param_list in self.__usage_paramiter_examples:
                self.__help_string += f"\n    {self.__command_name if self.__command_name is not None else f'python {self.__file_name}'} {param_list}"

    @property
    def help_string(self):
        return self.__help_string

    def parse_arguments(self, arguments: Union[Sequence[str], None] = None):

        self.__parameters.reset() # Make sure all parameters are unaltered from their starting values

        # Select the parameter strings
        self.__raw_arguments = arguments if arguments is not None else sys.argv[1:]
        number_of_raw_args = len(self.__raw_arguments)

        # Create a lookup of named parameters
        named_argument_lookup: Dict[str, str] = {}
        for arg_index, raw_arg in enumerate(self.__raw_arguments):
            if len(raw_arg) == 1:
                continue # Argument too short to be a parameter name
            if raw_arg[0] == "-":
                # Argument starts with a dash, so could be a param name
                # CAREFULL! Could be a negitive number or some user defined string starting with a dash.
                if len(raw_arg) > 2 and raw_arg[:2] == "--":
                    # Highly likley to be a parameter and definately not a short name
                    if raw_arg[2:] in self.__parameter_lookup_by_name:
                        param_name = raw_arg[2:]
                        if isinstance(self.__parameter_lookup_by_name[param_name], self.Flag):
                            named_argument_lookup[param_name] = ""
                        else:
                            if arg_index + 1 == number_of_raw_args:
                                raise IndexError("Final argument was parameter name that required a value.")
                            named_argument_lookup[param_name] = self.__raw_arguments[arg_index + 1]
                else:
                    # Could still be a short name
                    if raw_arg[1:] in self.__parameter_lookup_by_short_name:
                        param_short_name = raw_arg[1:]
                        param = self.__parameter_lookup_by_short_name[param_short_name]
                        if isinstance(param, self.Flag):
                            named_argument_lookup[param.name] = ""
                        else:
                            if arg_index + 1 == number_of_raw_args:
                                raise IndexError("Final argument was parameter name that required a value.")
                            named_argument_lookup[param.name] = self.__raw_arguments[arg_index + 1]

        for param in self.__parameters:
            param.set_from_args(named_argument_lookup)
        positional_arg_index = 0
        for param in self.__parameters:
            if isinstance(param, self.PositionalParam):
                if not param.is_explicitly_set:
                    if len(self.__raw_arguments) <= positional_arg_index:
                        raise IndexError("Too few arguments provided for the number of positional parameters.")
                    param.set_from_positional_arg(self.__raw_arguments[positional_arg_index])
                    positional_arg_index += 1
            else:
                if param.name not in self.__parameters.special_param_names:
                    break # No more positional parameters
        
        if not self.__parameters.help_param.value:
            # If help flag has been set, skip validation as target function won't actually be called.
            for param in self.__parameters:
                param.validate()
        self.__arguments_have_been_parsed = True

    @property
    def stopwatch(self) -> Stopwatch:
        self.__stopwatch.print_start
        return self.__stopwatch

    @staticmethod
    def __run(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            is_root_process = True
            try:
                if not self.__arguments_have_been_parsed:
                    self.parse_arguments() # Automatically parse command line arguments

                if not self.use_MPI or get_mpi_rank() == 0:
                    # Either the first MPI process or MPI isn't active
                    self.__stopwatch.print_start()
                    Console.reset_stopwatch(self.__stopwatch)
                    result = func(self, *args, **kwargs)
                    self.__stopwatch.print_stop()

                else:
                    if not _settings_object.mpi_avalible:
                        raise ImportError("MPI required but not available.")
                    # MPI active and not the first process
                    is_root_process = False
                    self.__stopwatch.start()
                    Console.reset_stopwatch(self.__stopwatch)
                    result = func(self, *args, **kwargs)
                    self.__stopwatch.stop()

                return result
                    
            except Exception as e:
                if is_root_process or self.MPI_print_non_root_process_errors:
                    has_message = str(e) != ""
                    Console.print_error(f"Execution encountered an error{(':' if has_message else ' (no details available).') if _settings_object.debug or _settings_object.verbose else '.'}")
                    Console.print_debug("Traceback (most recent call last):\n" + "".join(traceback.format_tb(e.__traceback__)) + type(e).__name__ + (f": {str(e)}" if has_message else ""))
                    if has_message and not _settings_object.debug:
                        Console.print_error(str(e))
                    Console.print_info("Terminating...")
                    self.stopwatch.print_stop()

        return inner

    @__run
    def run(self, func):
        if self.__parameters.help_param.value:
            print(self.help_string)
            sys.exit()
        if self.__parameters.debug_param.value:
            _settings_object.enable_debug()
        if self.__parameters.verbose_param.value:
            _settings_object.enable_verbose()
        kwargs = self.__parameters.get_keyword_arguments()
        return func(**kwargs)

    @__run
    def run_with_async(self, func):
        if self.__parameters.help_param.value:
            print(self.help_string)
            sys.exit()
        if self.__parameters.debug_param.value:
            _settings_object.enable_debug()
        if self.__parameters.verbose_param.value:
            _settings_object.enable_verbose()
        kwargs = self.__parameters.get_keyword_arguments()
        return start_main_async(func, **kwargs)

#    @staticmethod
#    def passthrough_converter(value: str) -> str:
#        return value
#
#    @staticmethod
#    def bool_converter(value: str) -> bool:
#        return value.lower() in ("true", "t", "yes", "y", "1", "accept", "allow", "confirm")
#
#    @staticmethod
#    def filter_converter(value: str) -> Tuple[bool, ...]:
#        return tuple(ScriptWrapper.bool_converter(element) for element in value.lower())
#
#    #@staticmethod
#    #def reverse_flag_converter(value: bool) -> bool:
#    #    return not value
#
#    @staticmethod
#    def make_list_converter(seperator = ",", item_converters: Union[Sequence[Callable[[str], Any]], Callable[[str], Any], None] = None):
#        if item_converters is None:
#            item_converters = ScriptWrapper.passthrough_converter
#        def list_converter(value: str) -> Any:
#            if not isinstance(item_converters, Sequence):
#                return [item_converters(item) if item != "" or item_converters == ScriptWrapper.passthrough_converter or item_converters == str else None for item in value.split(seperator)]
#            else:
#                converters_for_use: List[Callable[[str], Any]] = list(item_converters)
#                strings = value.split(seperator)
#                while len(converters_for_use) < len(strings):
#                    converters_for_use.append(ScriptWrapper.passthrough_converter)
#                return [converters_for_use[i](strings[i]) if strings[i] != "" or converters_for_use[i] == ScriptWrapper.passthrough_converter or converters_for_use == str else None for i in range(len(strings))]
#        return list_converter
    passthrough_converter = ScriptWrapper_passthrough_converter
    bool_converter = ScriptWrapper_bool_converter
    filter_converter = ScriptWrapper_filter_converter
    #reverse_flag_converter = ScriptWrapper_reverse_flag_converter
    make_list_converter = ScriptWrapper_make_list_converter



'''
class ScriptWrapper(object):
    """
    Singelton that handles command argument parsing for the currently running program.
    Constructor Arguments:
              str filename                 --> The name of the entrypoint file.
              str author                   --> The name(s) of the file author(s).
              str version_string           --> String indicating the current version of the file.
              str edit_date                --> String indicating the date of the last edit.
              str description              --> A description of the function of the script.
        list[str] dependancies             --> A list of dependancies of the script.
        list[str] usage_paramiter_examples --> A list of strings demonstrating example paramiter
                                                   lists (exclude the call to the script).
             list positional_params        --> List of paramiter information for paramiters that must
                                                   appear at the start of the argument list in order
                                                   without a flag identifier (each paramiter is a list
                                                   with fields as below).
                                                   [
                                                    name, description, conversion func (nullable)
                                                   ]
             list params                   --> List of paramiter information (each paramiter is a list
                                                   with fields as below).
                                                   [
                                                    name, short name (nullable), description,
                                                    required?, flag?, conversion func (nullable),
                                                    default value, mutually exclusive flags (list[str])
                                                   ]
    Public Methods:
               run
        static passthrough_converter
        static bool_converter
        static make_list_converter
    """

    #TODO: set this up to enforce a singleton
    #__initialised = False
    #__instance = None
    #
    #def __new__(cls, *args, **kwargs):
    #    if ScriptWrapper.__initialised:
    #        #raise RuntimeError("A running program may only define one ScriptWrapper instance.")
    #        return ScriptWrapper.__instance
    #    else:
    #        ScriptWrapper.__initialised = True
    #        ScriptWrapper.__instance = super(ScriptWrapper, cls).__new__(cls)
    #        return ScriptWrapper.__instance

    def __init__(self, filename: str, author: str, version_string: str, edit_date: str,
                       description: str,
                       dependancies: list,
                       usage_paramiter_examples: list,
                       positional_params: list = [], params: list = []):
        self.raw_case_args = sys.argv[1:]
        self.lowercase_args = [arg.lower() for arg in sys.argv[1:]]
        if "-d" in self.lowercase_args or "--debug" in self.lowercase_args:
            _settings_object.enable_debug()
        if "-v" in self.lowercase_args or "--verbose" in self.lowercase_args:
            _settings_object.enable_verbose()

        rendered_dependancy_strings = "\n".join([f"    {dependancy}" for dependancy in dependancies])

        rendered_usage_examples = "\n".join([f"    python {filename} {param_list}" for param_list in usage_paramiter_examples])

        self.n_positionals = len(positional_params)
        
        self.argument_spec = {}
        self.argument_spec["help"] = { "short_name": "h", "description": "Display this docstring.", "is_required": False, "is_positional": False, "is_flag": True, "mutually_exclusive_flags": [], "converter_function": ScriptWrapper.passthrough_converter, "default_value": None  }
        self.argument_spec["debug"] = { "short_name": "d", "description": "Display debug infomation.", "is_required": False, "is_positional": False, "is_flag": True, "mutually_exclusive_flags": [], "converter_function": ScriptWrapper.passthrough_converter, "default_value": None  }
        self.argument_spec["verbose"] = { "short_name": "v", "description": "Display progression infomation.", "is_required": False, "is_positional": False, "is_flag": True, "mutually_exclusive_flags": [], "converter_function": ScriptWrapper.passthrough_converter, "default_value": None  }
        for param in positional_params:
            self.argument_spec[param[0]] = { "short_name": None, "description": param[1], "is_required": True, "is_positional": True, "is_flag": False, "converter_function": ScriptWrapper.passthrough_converter if param[2] is None else param[2], "default_value": None }
        for param in params:
            self.argument_spec[param[0]] = { "short_name": param[1], "description": param[2], "is_required": param[3], "is_positional": False, "is_flag": param[4], "mutually_exclusive_flags": [] if len(param) <= 7 else param[7], "converter_function": ScriptWrapper.passthrough_converter if param[5] is None else param[5], "default_value": param[6] if param[6] is not None else ((ScriptWrapper.passthrough_converter if param[5] is None else param[5])(False) if param[4] else None) }

        name_spacing = max([len(name) for name in self.argument_spec])
        indent_spacing = name_spacing + 26

        rendered_argument_spec = "".join(["\n\n    ({}) {}{}{} || {} --> {}".format((" p" if self.argument_spec[name]["is_positional"] else (" r" if self.argument_spec[name]["is_required"] else ("f " if self.argument_spec[name]["is_flag"] else "f+"))), "  " if self.argument_spec[name]["is_positional"] else "--", name, (name_spacing - len(name)) * " ", ("  " if self.argument_spec[name]["short_name"] is None else "-{}".format(self.argument_spec[name]["short_name"])), self.argument_spec[name]["description"].replace("\n", "\n" + (indent_spacing * " "))) for name in self.argument_spec])

        self.help_string = f"""
File: {filename}
Author: {author}
Vesion: {version_string}
Date:   {edit_date}
{description}
Commandline arguments & flags ( p = required positional argument,
                                r = required paramiter,
                               f  = optional flag,
                               f+ = optional flag with a required argument):{rendered_argument_spec}
Dependancies:
{rendered_dependancy_strings}
Example Usage:
{rendered_usage_examples}
"""

        if "-h" in self.lowercase_args or "--help" in self.lowercase_args:
            print(self.help_string)
            sys.exit()

        self.params = {}
        for name in self.argument_spec:
            self.params[name] = self.argument_spec[name]["default_value"] 

        param_check_strings = []
        param_name_reverse_lookup = {}
        for name in self.argument_spec:
            param_check_strings.append(f"--{name}")
            if self.argument_spec[name]["short_name"] is not None:
                param_check_strings.append("-{}".format(self.argument_spec[name]["short_name"]))
                param_name_reverse_lookup[self.argument_spec[name]["short_name"]] = name

        arg_index = 0
        while arg_index < len(self.lowercase_args):
            if self.lowercase_args[arg_index] in param_check_strings:
                param_name = None
                if self.lowercase_args[arg_index][1] == "-":
                    param_name = self.lowercase_args[arg_index][2:]
                else:
                    param_name = param_name_reverse_lookup[self.lowercase_args[arg_index][1:]]

                if self.argument_spec[param_name]["is_flag"]:
                    self.params[param_name] = self.argument_spec[param_name]["converter_function"](True)
                    for flag_name in self.argument_spec[param_name]["mutually_exclusive_flags"]:
                        self.params[flag_name] = self.argument_spec[param_name]["converter_function"](False)
                else:
                    self.params[param_name] = self.argument_spec[param_name]["converter_function"](self.raw_case_args[arg_index + 1])
                    arg_index += 1

            arg_index += 1

        for i in range(self.n_positionals):
            self.params[positional_params[i][0]] = self.argument_spec[positional_params[i][0]]["converter_function"](self.raw_case_args[i])

        self.use_MPI = False
        self.MPI_print_non_root_process_errors = False

    def __run(func):
        def inner(self, *args, **kwargs):
            is_root_process = True
            try:
                if not self.use_MPI or get_mpi_rank() == 0:
                    # Either the first MPI process or MPI isn't active
                    return func(self, *args, **kwargs)
                else:
                    if not _settings_object.mpi_avalible:
                        raise ImportError("MPI required but not avalible.")

                    # MPI active and not the first process
                    is_root_process = False
                    return func(self, *args, **kwargs)
                    
            except Exception as e:
                if is_root_process or self.MPI_print_non_root_process_errors:
                    has_message = e.__str__() != ""
                    Console.print_error(f"Execution encountered an error{(':' if has_message else ' (no details avalible).') if _settings_object.debug or _settings_object.verbose else '.'}")
                    Console.print_debug("Traceback (most recent call last):\n" + "".join(traceback.format_tb(e.__traceback__)) + type(e).__name__ + (f": {e.__str__()}" if has_message else ""))
                    if has_message and not _settings_object.debug:
                        Console.print_error(e.__str__())
                    Console.print_info("Terminating.")

        return inner

    @__run
    def run(self, func):
        kwargs = { param.replace("-", "_"): self.params[param] for param in self.params if param not in ("help", "debug", "verbose") }
        return func(**kwargs)

    @__run
    def run_with_async(self, func):
        kwargs = { param.replace("-", "_"): self.params[param] for param in self.params if param not in ("help", "debug", "verbose") }
        return start_main_async(func, **kwargs)

    @staticmethod
    def passthrough_converter(value: str) -> str:
        return value

    @staticmethod
    def bool_converter(value: str) -> bool:
        return value in ("true", "t", "yes", "y", "accept", "1")

    @staticmethod
    def reverse_flag_converter(value: bool) -> bool:
        return not value

    @staticmethod
    def make_list_converter(seperator = ",", item_converters: list = None):
        if item_converters is None: item_converters = ScriptWrapper.passthrough_converter
        def list_converter(value: str) -> bool:
            if not isinstance(item_converters, (list, tuple)):
                return [item_converters(item) if item != "" or item_converters == ScriptWrapper.passthrough_converter or item_converters == str else None for item in value.split(seperator)]
            else:
                strings = value.split(seperator)
                while len(item_converters) < len(strings):
                    item_converters.append(ScriptWrapper.passthrough_converter)
                return [item_converters[i](strings[i]) if strings[i] != "" or item_converters[i] == ScriptWrapper.passthrough_converter or item_converters == str else None for i in range(len(strings))]
        return list_converter
'''



from .__about__ import __version__

__doc__ = f"""
~| Quasar Code - Python Package |~

Adds functionality to Python along with wrapers for existing modules for ease of use

Credits:
    Written by Christopher Rowe
    Notable contribusions taken from code written by Tim Greenshaw 10/2018

Version: {__version__}
"""

from ._global_settings import settings_object as Settings

from . import edp

from . import Games

from . import IO

from . import MPI

from . import Science

from . import Plotting

from . import Tools

from .Tools._directorys_and_imports import source_file_relitive_add_to_path
from .IO.Text import Console, Stopwatch
from .edp import Event
from .Tools._async import start_main_async



from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Any, Generic, TypeVar, ParamSpec, Concatenate, Iterable, TypeVarTuple, Callable
from numbers import Number, Real

import numpy as np
from scipy.stats import norm

from ..Tools._Struct import CacheableStruct
from ..Tools._autoproperty import AutoProperty_NonNullable

T = TypeVar("T", bound = "CacheableFunction")
P = ParamSpec("P")
R = TypeVar("R")

class CacheableFunction(CacheableStruct, Generic[P, R], ABC, Callable[P, R]):
    constants = AutoProperty_NonNullable[dict[str, Any]]() # P.kwargs
    def __init__(self, cacheable_attributes: tuple[str, ...]|None = None, **kwargs) -> None:
        super().__init__(
            cacheable_attributes = (*(cacheable_attributes if cacheable_attributes is not None else tuple()), "constants"),
            **kwargs
        )
    @classmethod
    def new(cls: type[T], **constants: P.kwargs) -> T:
        """
        Create a new instance of the function using the provided constants.
        """
        instance = cls()
        instance.constants = constants
        return instance
    def __call__(self, *args: P.args) -> R:
        return self.function(*args, **self.constants)
    @abstractmethod
    def function(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """
        The functional form to be implemented by the subclass.
        """
        raise NotImplementedError("Subclasses must implement the 'function' method.")

class PolynomialFunction(CacheableFunction[[Real|int|float], Real|int|float]):
    @classmethod
    def new(cls, *constant_terms: Real|int|float) -> "PolynomialFunction": # type: ignore[override]
        """
        Create a new instance of the function using the provided constants.
        """
        return super().new(**{f"{i}": constant_terms[i] for i in range(len(constant_terms))})
    def function(self, x: Real|int|float, **constants: Real|int|float) -> Real|int|float:
        """
        Evaluate the polynomial function at a given point x using the provided constants.
        """
        number_of_terms = len(constants)
        return sum([x**(number_of_terms - 1 - i) * constants[f"{i}"] for i in range(number_of_terms)]) # type: ignore[return-value]

class LinearFunction(PolynomialFunction):
    @classmethod
    def new(cls, gradient: Real|int|float, intercept: Real|int|float) -> "LinearFunction": # type: ignore[override]
        """
        Create a new instance of the function using the provided constants.
        """
        return super().new(gradient, intercept) # type: ignore[return-value]

A = TypeVar("A", Real|int|float, np.ndarray[Any, np.dtype[np.integer|np.floating]])
class GaussianFunction(CacheableFunction[[A], A]):
    @classmethod
    def new(cls, mean: Real|int|float, standard_deviation: Real|int|float, height_multiplier: Real|int|float) -> "GaussianFunction": # type: ignore[override]
        """
        Create a new instance of the function using the provided constants.
        """
        return super().new(mean = mean, standard_deviation = standard_deviation, height_multiplier = height_multiplier)
    def function(self, x: A, mean: Real|int|float, standard_deviation: Real|int|float, height_multiplier: Real|int|float) -> A:
        """
        Evaluate the Gaussian function at a given point x using the provided constants.
        """
        return norm(loc = mean, scale = standard_deviation).pdf(x) * height_multiplier



from ... import Settings
from ...MPI import get_mpi_rank
import os
import sys
from abc import ABC, abstractmethod
import uuid

class CheckpointBase(ABC):
    def __init__(self, filepath: str, checkpoint_file_extension = "checkpoint", backup_file_extension = "backup"):
        self.__filepath = filepath
        self.__file_extension = checkpoint_file_extension if not Settings.mpi_avalible else f"{checkpoint_file_extension}.{get_mpi_rank()}"
        self.__backup_extension = f"{checkpoint_file_extension}.{backup_file_extension}" if not Settings.mpi_avalible else f"{checkpoint_file_extension}.{backup_file_extension}.{get_mpi_rank()}"

        self.__checkpoint_file = f"{self.__filepath}.{self.__file_extension}"
        self.__backup_file = f"{self.__filepath}.{self.__backup_extension}"

        self._update_file_avalibility()

    @property
    def checkpoint_file(self):
        return self.__checkpoint_file

    @property
    def backup_file(self):
        return self.__backup_file

    @property
    def checkpoint_avalible(self):
        return  self.__checkpoint_avalible

    @property
    def backup_avalible(self):
        return  self.__checkpoint_avalible

    def _update_file_avalibility(self):
        self.__checkpoint_avalible = os.path.exists(self.__checkpoint_file)
        self.__backup_avalible = os.path.exists(self.__backup_file)

    @abstractmethod
    def _read_file(self, filepath):
        raise NotImplementedError("Must be overridden by a child class.")

    def read_checkpoint(self):
        if not self.__checkpoint_avalible:
            raise FileNotFoundError("There is no current checkpoint to read.")
        return self._read_file(self.__checkpoint_file)

    def read_backup(self):
        if not self.__backup_avalible:
            raise FileNotFoundError("There is no backup checkpoint to read.")
        return self._read_file(self.__backup_file)

    @abstractmethod
    def _write_file(self, filepath, data, *args, **kwargs):
        raise NotImplementedError("Must be overridden by a child class.")

    def write_checkpoint(self, data, *args, **kwargs):
        if self.__checkpoint_avalible:
            self._move_to_backup()
        return self._write_file(self.__checkpoint_file, data, *args, **kwargs)
        self._update_file_avalibility()

    def _move_to_backup(self):
        if not self.__checkpoint_avalible:
            raise FileNotFoundError("No checkpoint to backup.")

        tempfile_exists = False
        if self.__backup_avalible:
            tempfile_name = f".QCtempcheckpointbackupfile-{str(uuid.uuid4())}"
            os.rename(self.__backup_file, tempfile_name)
            tempfile_exists = True

        self._update_file_avalibility()
        try:
            os.rename(self.__checkpoint_file, self.__backup_file)
            self._update_file_avalibility()
            if self.__checkpoint_avalible:
                raise FileExistsError("Failed to create backup from existing checkpoint.")
        except:
            if tempfile_exists:
                os.rename(tempfile_name, self.__backup_file)
                tempfile_exists = False
        finally:
            if tempfile_exists:
                os.remove(tempfile_name)
            self._update_file_avalibility()

    def remove_checkpoint(self):
        os.remove(self.__checkpoint_file)
        self._update_file_avalibility()

    def remove_backup(self):
        os.remove(self.__backup_file)
        self._update_file_avalibility()



from ._IPlayingCard import IPlayingCard
from enum import Enum

class PlayingCard(IPlayingCard):
    """
    A card from a standard deck of 52 playing cards

    Atributes:
        PlayingCard.AllowedValues Value --> The card's value
        PlayingCard.AllowedSuits Suit --> The card's suit
    """

    class AllowedValues(Enum):
        """
        Enum of card values from ace to king and joker
        """
        
        A = 0,# Ace

        Two = 1,# Two

        Three = 2,# Three

        Four = 3,# Four

        Five = 4,# Five

        Six = 5,# Six

        Seven = 6,# Seven

        Eight = 7,# Eight

        Nine = 8,# Nine

        Ten = 9,# Ten

        J = 10,# Jack

        Q = 11,# Queen

        K = 12,# King

        Jo = 13# Joker

    class AllowedSuits(Enum):
        """
        Enum of card suits - hearts, clubs, dimonds, spades, joker
        """
        
        H = 0,# Heart

        C = 1,# Club

        D = 2,# Dimond

        S = 3,# Spade

        J = 4# Joker

    def __init__(self, value: AllowedValues, suit: AllowedSuits):
        self.__value = value
        self.__suit = suit

    @property
    def Value(self) -> AllowedValues:
        """
        The card's value
        """
        return self.__value

    @property
    def Suit(self) -> AllowedSuits:
        """
        The card's suit
        """
        return self.__suit

    __ValueStrings = ("A", "2", "3", "4", "5", "6", "7", "8", "9", "10", "J", "Q", "K")

    def GetValueString(value: AllowedValues) -> str:
        """
        Retrive the string counterpart to the AllowedValues enum value
        """
        return PlayingCard.__ValueStrings[int(value)]

    __SuitStrings = ("Heart", "Club", "Dimond", "Spade", "Joker")

    def GetSuitString(suit: AllowedSuits) -> str:
        """
        Retrive the string counterpart to the AllowedSuits enum suit
        """
        return PlayingCard.__SuitStrings[int(suit)]


from QuasarCode.Tools import ScriptWrapper

class Test_ArrayVisuliser(object):
    def test_standard(self):

        def target_function(test_value: int, test_value_2: float, test_value_3: float, test_value_4: float, test_flag: bool, test_flag_2: bool):
            assert test_value == 12345
            assert test_value_2 == 0.12345
            assert test_value_3 == 1.2345
            assert test_value_4 == 12.345
            assert test_flag
            assert not test_flag_2

        wrapper = ScriptWrapper(
            version = "1.0.0",
            authors = [ScriptWrapper.AuthorInfomation(given_name = "Joe", family_name = "Bloggs")],
            description = "A test function.",
            parameters = [
                ScriptWrapper.PositionalParam[int]("test-value", None, None, int, description = "A test parameter."),
                ScriptWrapper.PositionalParam[float]("test-value-2", None, None, float, description = "A test parameter."),
                ScriptWrapper.RequiredParam[float]("test-value-3", "t3", None, float, description = "A test parameter."),
                ScriptWrapper.OptionalParam[float]("test-value-4", "4", None, float, description = "A test parameter."),
                ScriptWrapper.Flag("test-flag", "f", None, description = "A test parameter."),
                ScriptWrapper.Flag("test-flag-2", "f2", None, inverted = True, description = "A test parameter.")
            ]
        )

        wrapper.parse_arguments(["12345", "-t3", "1.2345", "-f", "-4", "12.345", "--test-value-2", "0.12345", "-f2"])

        wrapper.run(target_function)
        


import pytest
from decimal import Decimal
import numpy as np
from matplotlib import pyplot as plt

##from QuasarCode.Science.graph import Figure
##from QuasarCode.Science.data import errorString, sigFig
#from QuasarCode.Science import Figure, error_string, sig_fig

from QuasarCode.Science import error_string, is_consistant, sig_fig, standard_form, standard_form_string, standard_form_string_if_nessessary, polygon_y_error_region, polygon_x_error_region

class TestClass_Data(object):
    def test_sig_fig(self):
        assert "436784329.8287460000000000000" == str(sig_fig(436784329.828746, 15)), "Rounding to 15 sig fig failed."
        assert "436784329.8287500000000000000" == str(sig_fig(436784329.828746, 14)), "Rounding to 14 sig fig failed."
        assert "436784329.8287000000000000000" == str(sig_fig(436784329.828746, 13)), "Rounding to 13 sig fig failed."
        assert "436784329.8290000000000000000" == str(sig_fig(436784329.828746, 12)), "Rounding to 12 sig fig failed."
        assert "436784329.8300000000000000000" == str(sig_fig(436784329.828746, 11)), "Rounding to 11 sig fig failed."
        assert "436784329.8000000000000000000" == str(sig_fig(436784329.828746, 10)), "Rounding to 10 sig fig failed."
        assert "436784330.0000000000000000000" == str(sig_fig(436784329.828746, 9)), "Rounding to 9 sig fig failed."
        assert "436784330.0000000000000000000" == str(sig_fig(436784329.828746, 8)), "Rounding to 8 sig fig failed."
        assert "436784300.0000000000000000000" == str(sig_fig(436784329.828746, 7)), "Rounding to 7 sig fig failed."
        assert "436784000.0000000000000000000" == str(sig_fig(436784329.828746, 6)), "Rounding to 6 sig fig failed."
        assert "436780000.0000000000000000000" == str(sig_fig(436784329.828746, 5)), "Rounding to 5 sig fig failed."
        assert "436800000.0000000000000000000" == str(sig_fig(436784329.828746, 4)), "Rounding to 4 sig fig failed."
        assert "437000000.0000000000000000000" == str(sig_fig(436784329.828746, 3)), "Rounding to 3 sig fig failed."
        assert "440000000.0000000000000000000" == str(sig_fig(436784329.828746, 2)), "Rounding to 2 sig fig failed."
        assert "400000000.0000000000000000000" == str(sig_fig(436784329.828746, 1)), "Rounding to 1 sig fig failed."

    def test_standardForm(self):
        print(standard_form_string(31415926.535, 11))
        result_num, result_power = standard_form(31415926.535)
        assert sig_fig(3.1415926535, 11) == sig_fig(result_num, 11)
        assert 7 == result_power

        print(standard_form_string(0.000031415926535, 11, latex = True))
        result_num, result_power = standard_form(0.000031415926535)
        assert sig_fig(3.1415926535, 11) == sig_fig(result_num, 11)
        assert -5 == result_power

        print(standard_form_string_if_nessessary(10543.2, 6))
        assert standard_form_string_if_nessessary(10543.2, 6) == "1.05432 * 10^4"

        print(standard_form_string_if_nessessary(1054.32, 6))
        assert standard_form_string_if_nessessary(1054.32, 6) == "1054.32"

        print(standard_form_string_if_nessessary(10.5432, 6))
        assert standard_form_string_if_nessessary(10.5432, 6) == "10.5432"

    def test_trueRound(self):
        pass# This is covered by the test_sig_fig function

    def test_consistancyTest(self):
        assert not is_consistant(0.0, 5.0, 1.0, 3.0, 1)
        assert is_consistant(0.0, 5.0, 1.0, 3.0, 2)
        assert not is_consistant(-5.0, 5.0, 7.0, 7.0, 1)
        assert is_consistant(-5.0, 5.0, 11.0, 11.0, 1)

    def test_errorString(self):
        assert "2.0435922040733323 \u00B1 0.7" == error_string(2.0435922040733323, 0.7)
        assert "0.7266078556762237 \u00B1 3.0" == error_string(0.7266078556762237, 3.0)

 #   def test_GraphAndErrorString(self):
 #       x = np.array([1.3, 2.1, 2.8, 3.9, 5.2])
 #       y = np.array([3.4, 4.6, 7.2, 8.5, 11.2])
 #
 #       dx = np.array([1, 1, 1, 1, 1])
 #       dy = np.array([1, 1, 1, 1, 1])
 #
 #       testGraph = Figure(x, y, dx, dy, size = None, title = "A test graph", xLabel = "X-Axis", yLabel = "Y-Axis", saveName = "testingFigure.png")
 #       testGraph.createFitLine()
 #       testGraph.plot()
 #       testGraph.save()
 #       #testGraph.show()# TODO: window hangs
 #
 #       m = testGraph.fit.fitVariables["m"]
 #       c = testGraph.fit.fitVariables["c"]
 #
 #       assert "2.0435922040733323" == str(m[0]), "The value calculated for the gradient does not match the expected value."
 #       assert "0.7481821730455976" == str(m[1]), "The error calculated for the gradient does not match the expected value."
 #       assert "0.7266078556762237" == str(c[0]), "The value calculated for the intercept does not match the expected value."
 #       assert "2.505349774682801" == str(c[1]), "The error calculated for the intercept does not match the expected value."
 #
 #       assert "2.0435922040733323 \u00B1 0.7" == error_string(m[0], m[1]), "The normalised value and error calculated for the gradient does not match the expected value."
 #       assert "0.7266078556762237 \u00B1 3.0" == error_string(c[0], c[1]), "The normalised value and error calculated for the intercept does not match the expected value."


class TestClass_Plotting(object):
    def test_plotting(self):
        plt.figure()
        polygon_y_error_region([0, 1, 2, 3, 4, 5], [1, 3, 1, 3, 1, 3], [-3, -1, -3, -1, -3, -1], "default", color = "red", alpha = 0.5)
        polygon_y_error_region([0, 1, 2, 3, 4, 5], [3, 1, 3, 1, 3, 1], [-1, -3, -1, -3, -1, -3], plt.gca(), color = "yellow", alpha = 0.5)
        plt.xlim((0, 5))
        plt.ylim((-3, 3))
        plt.show()

        plt.figure()
        polygon_x_error_region([-3, -1, -3, -1, -3, -1], [1, 3, 1, 3, 1, 3], [0, 1, 2, 3, 4, 5], "default", color = "red", alpha = 0.5)
        polygon_x_error_region([-1, -3, -1, -3, -1, -3], [3, 1, 3, 1, 3, 1], [0, 1, 2, 3, 4, 5], plt.gca(), color = "yellow", alpha = 0.5)
        plt.xlim((-3, 3))
        plt.ylim((0, 5))
        plt.show()



import time
from abc import ABC, abstractmethod



class StopwatchNotStartedError(Exception):
    pass
class StopwatchStartedError(Exception):
    pass
class StopwatchStoppedError(Exception):
    pass



class IStopwatch(ABC):
    """
    Interface for stopwatch types.
    """

    @staticmethod
    @abstractmethod
    def start_new(*args, **kwargs) -> "IStopwatch":
        raise NotImplementedError()
    @staticmethod
    @abstractmethod
    def synchronise_start_new(synchronise: "IStopwatch", *args, **kwargs) -> "IStopwatch":
        raise NotImplementedError()
    
    @staticmethod
    @abstractmethod
    def new_group(number: int, *args, **kwargs) -> "tuple[IStopwatch]":
        raise NotImplementedError()
    @staticmethod
    @abstractmethod
    def start_new_group(number: int, *args, **kwargs) -> "tuple[IStopwatch]":
        raise NotImplementedError()
    @staticmethod
    @abstractmethod
    def synchronise_start_new_group(number: int, synchronise: "IStopwatch", *args, **kwargs) -> "tuple[IStopwatch]":
        raise NotImplementedError()

    @abstractmethod
    def reset(self) -> None:
        """
        Reset the stopwatch state.
        """
        raise NotImplementedError()
    @abstractmethod
    def start(self) -> None:
        raise NotImplementedError()
    @abstractmethod
    def synchronise_start(self, synchronise: "IStopwatch") -> None:
        raise NotImplementedError()
    @abstractmethod
    def lap(self) -> float:
        """
        Record the time of the call and returns the number of seconds since the last `lap` or `start` call.
        """
        raise NotImplementedError()
    @abstractmethod
    def stop(self) -> float:
        """
        Stop the stopwatch.
        Record the time of the call and returns the number of seconds since the last `lap` or `start` call.
        """
        raise NotImplementedError()
    @abstractmethod
    def get_start_time(self) -> float:
        raise NotImplementedError()
    @abstractmethod
    def get_elapsed_time(self) -> float:
        """
        Return the time in seconds since the stopwatch was started.
        """
        raise NotImplementedError()
    @abstractmethod
    def get_elapsed_time_lap(self) -> float:
        """
        Return the time in seconds since the last lap or stopwatch was started.
        """
        raise NotImplementedError()
    @abstractmethod
    def get_elapsed_time_last_lap(self) -> float:
        """
        Return the time in seconds of the last complete lap.

        Returns 0.0 if neither `lap` or `stop` have been called.
        """
        raise NotImplementedError()

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError()
    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError()
    @abstractmethod
    def __float__(self) -> float:
        raise NotImplementedError()
    @abstractmethod
    def __int__(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def running(self) -> bool:
        raise NotImplementedError()
    @property
    @abstractmethod
    def started(self) -> bool:
        raise NotImplementedError()
    @property
    @abstractmethod
    def stopped(self) -> bool:
        raise NotImplementedError()

    @property
    @abstractmethod
    def start_time(self) -> float|None:
        raise NotImplementedError()
    @property
    @abstractmethod
    def lap_times(self) -> tuple[float]:
        raise NotImplementedError()
    @property
    @abstractmethod
    def stop_time(self) -> float|None:
        raise NotImplementedError()
    


class SimpleStopwatch(IStopwatch):
    """
    Basic stopwatch implementation for easy timing.
    """

    @staticmethod
    def start_new() -> "SimpleStopwatch":
        stopwatch = SimpleStopwatch()
        stopwatch.start()
        return stopwatch
    @staticmethod
    def synchronise_start_new(synchronise: IStopwatch) -> "SimpleStopwatch":
        stopwatch = SimpleStopwatch()
        stopwatch.synchronise_start(synchronise)
        return stopwatch
    
    @staticmethod
    def new_group(number: int) -> "tuple[SimpleStopwatch]":
        return tuple([SimpleStopwatch() for _ in range(number)])
    @staticmethod
    def start_new_group(number: int) -> "tuple[SimpleStopwatch]":
        stopwatches = tuple([SimpleStopwatch() for _ in range(number)])
        stopwatches[0].start()
        for i in range(1, number):
            stopwatches[i].synchronise_start(stopwatches[0])
        return stopwatches
    @staticmethod
    def synchronise_start_new_group(number: int, synchronise: "IStopwatch") -> "tuple[SimpleStopwatch]":
        return tuple([SimpleStopwatch.synchronise_start_new(synchronise) for _ in range(number)])

    def __init__(self):
        self.__running:   bool
        self.__lap_times: list[float]
        self.__stop_time: float|None

        self.reset()

    def reset(self) -> None:
        self.__running   = False
        self.__lap_times = []
        self.__stop_time = None

    def start(self) -> None:
        if self.__running:
            raise StopwatchStartedError("Attempted to restart stopwatch without first stopping or manually resetting.")
        self.reset()
        self.__running = True
        # Get the start time just before execution returns to the calling scope
        self.__lap_times.append(time.time())

    def synchronise_start(self, synchronise: IStopwatch) -> None:
        if self.__running:
            raise StopwatchStartedError("Attempted to restart stopwatch without first stopping or manually resetting.")
        self.reset()
        self.__running = True
        # Get the start time just before execution returns to the calling scope
        self.__lap_times.append(synchronise.get_start_time())

    def lap(self) -> float:
        # Ask for the time immediately after call
        t = time.time()
        # Now do validation
        if self.__stop_time is not None:
            raise StopwatchStoppedError("Unable to lap a stopped stopwatch.")
        elif not self.__running:
            raise StopwatchNotStartedError("Unable to lap a stopwatch that has not yet been started.")
        # Then do logic
        self.__lap_times.append(t)
        return t - self.__lap_times[-2]

    def stop(self) -> float:
        # Ask for the time immediately after call
        t = time.time()
        # Now do validation
        if self.__stop_time is not None:
            raise StopwatchStoppedError("Unable to stop an already stopped stopwatch.")
        elif not self.__running:
            raise StopwatchNotStartedError("Unable to stop a stopwatch that has not yet been started.")
        # Then do logic
        self.__running = False
        self.__lap_times.append(t)
        self.__stop_time = t
        return t - self.__lap_times[-2]
    
    def get_start_time(self) -> float:
        if not self.__running and self.__stop_time is None:
            raise StopwatchNotStartedError("Unable to get start time of a stopwatch that has not yet been started.")
        return self.__lap_times[0]
    
    def get_elapsed_time(self) -> float:
        t = time.time()
        if self.__running:
            return t - self.__lap_times[0]
        elif self.__stop_time is not None:
            return self.__stop_time - self.__lap_times[0]
        else:
            return 0.0
    
    def get_elapsed_time_lap(self) -> float:
        t = time.time()
        if self.__running:
            return t - self.__lap_times[-1]
        elif self.__stop_time is not None:
            return self.__stop_time - self.__lap_times[-1]
        else:
            return 0.0
    
    def get_elapsed_time_last_lap(self) -> float:
        if not self.started:
            raise StopwatchNotStartedError("Unable to get lap information from a stopwatch that has not yet been started.")
        if len(self.__lap_times) < 2:
            return 0.0
        else:
            return self.__lap_times[-1] - self.__lap_times[-2]

    def __str__(self) -> str:
        t = time.time()
        return str(t - self.__lap_times[-1])

    def __repr__(self) -> str:
        t = time.time()
        if self.__running:
            return f"SimpleStopwatch: {t - self.__lap_times[0]}"
        elif self.__stop_time is not None:
            return f"SimpleStopwatch: {self.__stop_time - self.__lap_times[0]} (stopped)"
        else:
            return "SimpleStopwatch: not yet started"
        
    def __float__(self):
        return self.get_elapsed_time()
        
    def __int__(self):
        return int(self.get_elapsed_time())

    @property
    def running(self) -> bool:
        return self.__running
    @property
    def started(self) -> bool:
        return self.__running or self.__stop_time is not None
    @property
    def stopped(self) -> bool:
        return self.__stop_time is not None

    @property
    def start_time(self) -> float|None:
        if not self.started:
            return None
        else:
            return self.__lap_times[0]
    @property
    def lap_times(self) -> tuple[float]:
        if not self.started:
            return tuple()
        elif not self.stopped:
            return tuple(self.__lap_times[1:])
        else:
            return tuple(self.__lap_times[1:-1])
    @property
    def stop_time(self) -> float|None:
        if not self.stopped:
            return None
        else:
            return self.__lap_times[-1]



from typing import Union, List, Tuple, Dict, cast

from matplotlib.collections import PathCollection, PolyCollection
from matplotlib.contour import QuadContourSet
from matplotlib.lines import Line2D
import numpy as np
import matplotlib.pyplot as plt

from QuasarCode.Data import Rect
from QuasarCode.Plotting import CachedPlotFactory, CachedPlot, CachedPlotLine, CachedPlotScatter, CachedPlotErrorbar, CachedPlotHexbin, CachedPlotContour, Hexbin, Contour
from QuasarCode.IO.Caching import CacheTargetFactory, CacheTarget

class Test_CachedPlot(object):

    def test_line(self):

        plot_factory = CachedPlotFactory(".")
        assert isinstance(plot_factory, CachedPlotFactory)

        plot_data = plot_factory.new(Rect.create_from_size(0, 0, 10, 10))
        assert isinstance(plot_data, CachedPlot)
        assert plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert plot_data.flip_x is False

        cache_factory = CacheTargetFactory("test_cache/Test_CachedPlot/test_line/{file}.pickle", "file")

        cache = cache_factory.new(file = "test_line")

        line_coords = (
            [0, 1, 2, 3, 4, 5, 6, 7, 8,  9, 10],
            [0, 2, 1, 4, 3, 6, 5, 8, 7, 10,  9]
        )

        fig = plt.figure()
        ax = fig.gca()
        mplt_line_object = plt.plot(*line_coords, label = "Line", color = "blue", linestyle = "-")[0]
        plt.legend()

        plot_data.add_element("line", CachedPlotLine.from_line(mplt_line_object))
        plot_data.render(fig, ax)

        re_rendered_test_line: Line2D = plot_data.plot_elements["line"].result
        assert np.all(re_rendered_test_line.get_xydata() == mplt_line_object.get_xydata())

        plot_factory.save(plot_data, cache)

        loaded_plot_data: CachedPlot = plot_factory.load(cache)
        assert isinstance(loaded_plot_data, CachedPlot)
        assert loaded_plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert loaded_plot_data.flip_x is False

        loaded_plot_data.render(fig, ax)
        loaded_re_rendered_test_line: Line2D = loaded_plot_data.plot_elements["line"].result
        assert np.all(loaded_re_rendered_test_line.get_xydata() == mplt_line_object.get_xydata())

    def test_scatter(self):

        plot_factory = CachedPlotFactory(".")
        assert isinstance(plot_factory, CachedPlotFactory)

        plot_data = plot_factory.new(Rect.create_from_size(0, 0, 10, 10))
        assert isinstance(plot_data, CachedPlot)
        assert plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert plot_data.flip_x is False

        cache_factory = CacheTargetFactory("test_cache/Test_CachedPlot/test_scatter/{file}.pickle", "file")

        cache = cache_factory.new(file = "test_scatter")

        line_coords = (
            [0, 1, 2, 3, 4, 5, 6, 7, 8,  9, 10],
            [0, 2, 1, 4, 3, 6, 5, 8, 7, 10,  9]
        )

        fig = plt.figure()
        ax = fig.gca()
        mplt_scatter_object = plt.scatter(*line_coords, label = "Scatter", color = "blue", marker = "^")
        plt.legend()

        plot_data.add_element("scatter", CachedPlotScatter.from_points(mplt_scatter_object))
        plot_data.render(fig, ax)

        re_rendered_test_scatter: PathCollection = plot_data.plot_elements["scatter"].result
        assert np.all(re_rendered_test_scatter.get_offsets() == mplt_scatter_object.get_offsets())

        plot_factory.save(plot_data, cache)

        loaded_plot_data: CachedPlot = plot_factory.load(cache)
        assert isinstance(loaded_plot_data, CachedPlot)
        assert loaded_plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert loaded_plot_data.flip_x is False

        loaded_plot_data.render(fig, ax)
        loaded_re_rendered_test_scatter: PathCollection = loaded_plot_data.plot_elements["scatter"].result
        assert np.all(loaded_re_rendered_test_scatter.get_offsets() == mplt_scatter_object.get_offsets())

    def test_hexbin(self):

        plot_factory = CachedPlotFactory(".")
        assert isinstance(plot_factory, CachedPlotFactory)

        plot_data = plot_factory.new(Rect.create_from_size(0, 0, 10, 10))
        assert isinstance(plot_data, CachedPlot)
        assert plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert plot_data.flip_x is False

        cache_factory = CacheTargetFactory("test_cache/Test_CachedPlot/test_hex/{file}.pickle", "file")

        cache = cache_factory.new(file = "test_hex")

        coords = np.random.rand(1000, 2) * 10

        fig = plt.figure()
        ax = fig.gca()
        mplt_hex_object = plt.hexbin(coords[:, 0], coords[:, 1], cmap = "viridis", gridsize = 100, extent = plot_data.extent.extent)
        plt.legend()

        plot_data.add_element("hexbin", CachedPlotHexbin.from_hexes(mplt_hex_object, plot_data.extent, 100))
        plot_data.render(fig, ax)

        re_rendered_test_hexbin: PolyCollection = plot_data.plot_elements["hexbin"].result
        assert np.all(re_rendered_test_hexbin.get_offsets() == mplt_hex_object.get_offsets())
        assert np.all(re_rendered_test_hexbin.get_array() == mplt_hex_object.get_array())
        assert np.all(re_rendered_test_hexbin.get_alpha() == mplt_hex_object.get_alpha())

        plot_factory.save(plot_data, cache)

        loaded_plot_data: CachedPlot = plot_factory.load(cache)
        assert isinstance(loaded_plot_data, CachedPlot)
        assert loaded_plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert loaded_plot_data.flip_x is False

        loaded_plot_data.render(fig, ax)
        loaded_re_rendered_test_hexbin: PolyCollection = loaded_plot_data.plot_elements["hexbin"].result
        assert np.all(loaded_re_rendered_test_hexbin.get_offsets() == mplt_hex_object.get_offsets())
        assert np.all(loaded_re_rendered_test_hexbin.get_array() == mplt_hex_object.get_array())
        assert np.all(loaded_re_rendered_test_hexbin.get_alpha() == mplt_hex_object.get_alpha())

    def test_Hexbin(self):

        plot_factory = CachedPlotFactory(".")
        assert isinstance(plot_factory, CachedPlotFactory)

        plot_data = plot_factory.new(Rect.create_from_size(0, 0, 10, 10))
        assert isinstance(plot_data, CachedPlot)
        assert plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert plot_data.flip_x is False

        cache_factory = CacheTargetFactory("test_cache/Test_CachedPlot/test_hex/{file}.pickle", "file")

        cache = cache_factory.new(file = "test_hex_from_Hexbin")

        coords = np.random.rand(1000, 2) * 10

        hexbin_object = Hexbin(coords[:, 0], coords[:, 1])
        hexbin_object.plot_hexbin(extent = plot_data.extent, gridsize = 100, colourmap = "viridis")
        mplt_hex_object = hexbin_object.data.result
        assert isinstance(mplt_hex_object, PolyCollection)

        fig = plt.figure()
        ax = fig.gca()

        plot_data.add_element("hexbin", hexbin_object.data)
        plot_data.render(fig, ax)

        re_rendered_test_hexbin: PolyCollection = plot_data.plot_elements["hexbin"].result
        assert isinstance(re_rendered_test_hexbin, PolyCollection)
        assert np.all(re_rendered_test_hexbin.get_offsets() == mplt_hex_object.get_offsets())
        assert np.all(re_rendered_test_hexbin.get_array() == mplt_hex_object.get_array())
        assert np.all(re_rendered_test_hexbin.get_alpha() == mplt_hex_object.get_alpha())

        plot_factory.save(plot_data, cache)

        loaded_plot_data: CachedPlot = plot_factory.load(cache)
        assert isinstance(loaded_plot_data, CachedPlot)
        assert loaded_plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert loaded_plot_data.flip_x is False

        loaded_plot_data.render(fig, ax)
        loaded_re_rendered_test_hexbin: PolyCollection = loaded_plot_data.plot_elements["hexbin"].result
        assert isinstance(loaded_re_rendered_test_hexbin, PolyCollection)
        assert np.all(loaded_re_rendered_test_hexbin.get_offsets() == mplt_hex_object.get_offsets())
        assert np.all(loaded_re_rendered_test_hexbin.get_array() == mplt_hex_object.get_array())
        assert np.all(loaded_re_rendered_test_hexbin.get_alpha() == mplt_hex_object.get_alpha())

    def test_Contour(self):

        plot_factory = CachedPlotFactory(".")
        assert isinstance(plot_factory, CachedPlotFactory)

        plot_data = plot_factory.new(Rect.create_from_size(0, 0, 10, 10))
        assert isinstance(plot_data, CachedPlot)
        assert plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert plot_data.flip_x is False

        cache_factory = CacheTargetFactory("test_cache/Test_CachedPlot/test_hex/{file}.pickle", "file")

        cache = cache_factory.new(file = "test_hex_from_Hexbin")

        coords = np.random.rand(1000, 2) * 10

        contour_object = Contour(coords[:, 0], coords[:, 1])

        contour_object.generate(gridsize = 100) # Run this early to get the min/max levels

        contour_object.add_contours(
            (contour_object.max_level - contour_object.min_level) / 4,
            (contour_object.max_level - contour_object.min_level) / 2,
            (contour_object.max_level - contour_object.min_level) * 3/4
        )

        lvl0, lvl1, lvl2 = sorted(contour_object.contour_levels)
        assert lvl0 == (contour_object.max_level - contour_object.min_level) / 4
        assert lvl1 == (contour_object.max_level - contour_object.min_level) / 2
        assert lvl2 == (contour_object.max_level - contour_object.min_level) * 3/4

        contour_object.set_colour(lvl2, "black")
        contour_object.set_linestyle(lvl2, "-")

        contour_object.set_colour(lvl1, "blue")
        contour_object.set_linestyle(lvl1, "--")

        contour_object.set_colour(lvl0, "green")
        contour_object.set_linestyle(lvl0, ":")

        fig = plt.figure()
        ax = fig.gca()

        mplt_hex_object = contour_object.plot(ax)
        assert isinstance(mplt_hex_object, QuadContourSet)
        assert isinstance(contour_object.plotting_result, QuadContourSet)

        plot_data.add_element("contour", contour_object.data)
        plot_data.render(fig, ax)

        re_rendered_test_contour: QuadContourSet = plot_data.plot_elements["contour"].result
        assert isinstance(re_rendered_test_contour, QuadContourSet)
        assert np.all(re_rendered_test_contour.levels == mplt_hex_object.levels)
        assert np.all(re_rendered_test_contour.alpha == mplt_hex_object.alpha)
        assert np.all(re_rendered_test_contour.colors == mplt_hex_object.colors)
        assert np.all(re_rendered_test_contour.linestyles == mplt_hex_object.linestyles)
        assert np.all(re_rendered_test_contour.get_linewidth() == mplt_hex_object.get_linewidth())

        plot_factory.save(plot_data, cache)

        loaded_plot_data: CachedPlot = plot_factory.load(cache)
        assert isinstance(loaded_plot_data, CachedPlot)
        assert loaded_plot_data.extent == Rect.create_from_size(0, 0, 10, 10)
        assert loaded_plot_data.flip_x is False

        loaded_plot_data.render(fig, ax)
        loaded_re_rendered_test_contour: QuadContourSet = loaded_plot_data.plot_elements["contour"].result
        assert isinstance(loaded_re_rendered_test_contour, QuadContourSet)
        assert np.all(loaded_re_rendered_test_contour.levels == mplt_hex_object.levels)
        assert np.all(loaded_re_rendered_test_contour.alpha == mplt_hex_object.alpha)
        assert np.all(loaded_re_rendered_test_contour.colors == mplt_hex_object.colors)
        assert np.all(loaded_re_rendered_test_contour.linestyles == mplt_hex_object.linestyles)
        assert np.all(loaded_re_rendered_test_contour.get_linewidth() == mplt_hex_object.get_linewidth())




from ..Tools._Struct import Struct
from ..Tools._autoproperty import TypedAutoProperty, NullableTypedAutoProperty
from ..Tools._typeshield import TypeShield

from typing import Union

class VersionInfomation_Delta(Struct):
    major    = TypedAutoProperty[int](TypeShield[int](int))
    minor    = TypedAutoProperty[int](TypeShield[int](int))
    revision = TypedAutoProperty[int](TypeShield[int](int))
    alpha    = TypedAutoProperty[int](TypeShield[int](int))
    beta     = TypedAutoProperty[int](TypeShield[int](int))

class VersionInfomation(Struct):
    major    = NullableTypedAutoProperty[Union[int, None]](TypeShield[Union[int, None]](int, type(None)), default_value = 0)
    minor    = NullableTypedAutoProperty[Union[int, None]](TypeShield[Union[int, None]](int, type(None)), default_value = 0)
    revision = NullableTypedAutoProperty[Union[int, None]](TypeShield[Union[int, None]](int, type(None)), default_value = 0)
    alpha    = NullableTypedAutoProperty[Union[int, None]](TypeShield[Union[int, None]](int, type(None)), default_value = 0)
    beta     = NullableTypedAutoProperty[Union[int, None]](TypeShield[Union[int, None]](int, type(None)), default_value = 0)

    @property
    def is_alpha(self) -> bool:
        return (not self.is_beta) and (self.alpha is not None) and (self.alpha > 0)

    @property
    def is_beta(self) -> bool:
        return (self.beta is not None) and (self.beta > 0)

    def to_string(self, minimum_length = False):
        result = ""
        if self.is_alpha:
            result = f"alpha{self.alpha}"
        elif self.is_beta:
            result = f"beta{self.beta}"
        if not minimum_length or len(result) > 0 or (self.revision is not None and self.revision > 0):
            result = f"{self.revision if self.revision is not None else 0}" + (f".{result}" if len(result) > 0 else "")
        if not minimum_length or len(result) > 0 or (self.minor is not None and self.minor > 0):
            result = f"{self.minor if self.minor is not None else 0}" + (f".{result}" if len(result) > 0 else "")
        result = f"{self.major if self.major is not None else 0}" + (f".{result}" if len(result) > 0 else "")

        return result

    def __str__(self):
        return self.to_string(minimum_length = False)
    
    @staticmethod
    def from_string(value: str) -> "VersionInfomation":
        elements = value.split(".")
        if len(elements) == 0:
            elements = ["0"]
        elif len([e for e in elements if e == ""]) > 0:
            raise ValueError(f"{value} is not a valid version number string (some components blank).")
        elif len(elements) > 4:
            raise ValueError(f"{value} is not a valid version number string (too many components).")
        converted_elements = [int(e) for e in elements[:3]]
        if len(converted_elements) < 3:
            converted_elements.extend([0] * (3 - len(converted_elements)))
        converted_elements.append(int(elements[3][5:]) if len(elements) == 4 and elements[3][:5] == "alpha" else 0)
        converted_elements.append(int(elements[3][4:]) if len(elements) == 4 and elements[3][:4] == "beta" else 0)
        return VersionInfomation(
               major = converted_elements[0],
               minor = converted_elements[1],
            revision = converted_elements[2],
               alpha = converted_elements[3],
                beta = converted_elements[4]
        )

    @staticmethod
    def from_int(value: int) -> "VersionInfomation":
        if value < 0:
            raise ValueError("Can't have a negitive version number.")
        return VersionInfomation(
               major = value,
               minor = 0,
            revision = 0,
               alpha = 0,
                beta = 0
        )
    
    def __eq__(self, value: object) -> bool:
        if isinstance(value, int):
            return self.major == value and self.minor in (0, None) and self.revision in (0, None) and self.alpha is None and self.beta is None
        elif isinstance(value, VersionInfomation):
            return self.major == value.major and self.minor == value.minor and self.revision == value.revision and self.alpha == value.alpha and self.beta == value.beta
        else:
            return False
        
    def __ne__(self, value: object) -> bool:
        return not self.__eq__(value)
    
    def __sub__(self, value: Union["VersionInfomation", int]) -> VersionInfomation_Delta:
        if isinstance(value, int):
            return VersionInfomation_Delta(
                   major = (self.major if self.major is not None else 0) - value,
                   minor = 0,
                revision = 0,
                   alpha = 0,
                    beta = 0
            )
        elif isinstance(value, VersionInfomation):
            return VersionInfomation_Delta(
                   major = (self.major    if self.major    is not None else 0) - (value.major    if value.major    is not None else 0),
                   minor = (self.minor    if self.minor    is not None else 0) - (value.minor    if value.minor    is not None else 0),
                revision = (self.revision if self.revision is not None else 0) - (value.revision if value.revision is not None else 0),
                   alpha = (self.alpha    if self.alpha    is not None else 0) - (value.alpha    if value.alpha    is not None else 0),
                    beta = (self.beta     if self.beta     is not None else 0) - (value.beta     if value.beta     is not None else 0)
            )
        else:
            raise TypeError(f"Unsupported operation between VersionInfomation and object of type {type(value)}.")
    
    def __add__(self, value: VersionInfomation_Delta) -> "VersionInfomation":
        if isinstance(value, int):
            if (self.major if self.major is not None else 0) + value < 0:
                raise ValueError(f"Unable to subtract {-value} from a major version of {(self.major if self.major is not None else 0)}.")
            
            return VersionInfomation(
                   major = (self.major if self.major is not None else 0) + value,
                   minor = 0,
                revision = 0,
                   alpha = 0,
                    beta = 0
            )
        
        elif isinstance(value, VersionInfomation):
            new_major    = (self.major    if self.major    is not None else 0) + value.major,
            new_minor    = (self.minor    if self.minor    is not None else 0) + value.minor,
            new_revision = (self.revision if self.revision is not None else 0) + value.revision,
            new_alpha    = (self.alpha    if self.alpha    is not None else 0) + value.alpha,
            new_beta     = (self.beta     if self.beta     is not None else 0) + value.beta

            if new_major < 0:    raise ValueError(f"Unable to subtract {-value} from a major version of {   (self.major    if self.major    is not None else 0)}.")
            if new_minor < 0:    raise ValueError(f"Unable to subtract {-value} from a minor version of {   (self.minor    if self.minor    is not None else 0)}.")
            if new_revision < 0: raise ValueError(f"Unable to subtract {-value} from a revision version of {(self.revision if self.revision is not None else 0)}.")
            if new_alpha < 0:    raise ValueError(f"Unable to subtract {-value} from a alpha version of {   (self.alpha    if self.alpha    is not None else 0)}.")
            if new_beta < 0:     raise ValueError(f"Unable to subtract {-value} from a beta version of {    (self.beta     if self.beta     is not None else 0)}.")

            return VersionInfomation_Delta(
                   major = new_major,
                   minor = new_minor,
                revision = new_revision,
                   alpha = new_alpha,
                    beta = new_beta
            )
        
        else:
            raise TypeError(f"Unsupported operation between VersionInfomation and object of type {type(value)}.")



import numpy as np
from typing import Union, Generic, TypeVar, Callable, Iterable, Sized, SupportsIndex, Collection, Sequence, Dict, Any, cast
from typing_extensions import ParamSpec, Concatenate
from unyt import unyt_array

P = ParamSpec("P")
T = TypeVar("T")
class Cast(Generic[T]):
    def __init__(self, cast_func: Callable[Concatenate[Any, P], T], *args: P.args, nullable: bool = False, **kwargs: P.kwargs) -> None:
        self.__target_func = cast_func
        self.__nullable = nullable
        self.__args = args
        self.__kwargs = kwargs

    @property
    def nullable(self) -> bool:
        return self.__nullable

    def __call__(self, value: Any) -> Union[T, None]:
        return None if value is None and self.__nullable else self.__target_func(value, *self.__args, **self.__kwargs)
    
    @staticmethod
    def passthrough() -> "Cast[Any]":
        return Cast(lambda o, *args, **kwargs: o, nullable = True)



def non_nullable_cast(cast_func: Callable[Concatenate[Any, P], T], *args: P.args, **kwargs: P.kwargs) -> Cast[T]:
    """
    Creates a wrapper function that allows other arguments to be passed to a cast function.
    None values are passed through to the function!
    """
    return Cast(cast_func, nullable = False, *args, **kwargs)

def nullable_cast(cast_func: Callable[Concatenate[Any, P], T], *args: P.args, **kwargs: P.kwargs) -> Cast[T]:
    """
    Creates a wrapper function that allows None values to bypass a cast function.

    value is None:     return None
    value is NOT None: apply cast
    """
    return Cast(cast_func, nullable = True, *args, **kwargs)

class NestedCast(Cast[T]):
    def __init__(self, *layer_cast_funcs: Union[Callable[..., Any], Cast], nullable: bool = False, layer_args: Union[Sequence[Union[Collection[Any], None]], None] = None, layer_kwargs: Union[Sequence[Union[Dict[str, Any], None]], None] = None) -> None:
        self.__target_layer_funcs = layer_cast_funcs
        self.__last_layer_index = len(self.__target_layer_funcs) - 1

        if layer_args is None:
            layer_args = [None for _ in range(len(self.__target_layer_funcs))]
        if layer_kwargs is None:
            layer_kwargs = [None for _ in range(len(self.__target_layer_funcs))]

        super().__init__(self._traverse_nest, layer_args, layer_kwargs)

    def _traverse_nest(self, value: Any, layer_args: Sequence[Union[Collection[Any], None]], layer_kwargs: Sequence[Union[Dict[str, Any], None]], layer: int = 0):
        if layer != self.__last_layer_index and isinstance(value, Sized) and isinstance(value, SupportsIndex):
            if isinstance(value, unyt_array):
                value = list(value.value) # The target must not have attached units as this might brake on looking at the values!
            if isinstance(value, np.ndarray):
                value = list(value) # The target must not be a ndarray as this will mutate the cast elements when returned to the target!
            if isinstance(value, tuple):
                value = list(value) # The target must be mutable!
            for i, inner_value in enumerate(value):
                value[i] = self._traverse_nest(inner_value, layer_args, layer_kwargs, layer = layer + 1)

        args = cast(Collection[Any], layer_args[layer] if layer_args[layer] is not None else [])
        kwargs = cast(Dict[str, Any], layer_kwargs[layer] if layer_kwargs[layer] is not None else {})
        return self.__target_layer_funcs[layer](value, *args, **kwargs)



from random import Random
from ._IPlayingCard import IPlayingCard

class CardGroup(list):
    """
    A list of IPlayingCards. Can be used as an ordenary Python list.

    Atributes:
        str Name --> name of the collection

    __init__():
        str name --> name of the collection
        list cards --> list of IPlayingCards to initialise the collection with (deafult is blank list)
    """
    def __init__(self, name: str, cards: list = []):
        self.Name = name
        self += cards

    def Remove(self, index: int) -> IPlayingCard:
        """
        Remove a card from the collection at the specified index
        """
        element = self[index]
        del self[index]
        return element

    def Shuffle(self):
        """
        Randomises the order of the cards in the group
        """
        generator = Random()
        newOrder = []

        while len(self) > 0:
            position = generator.randint(0, len(self) - 1)
            newOrder.append(self.Remove(position))

        self += newOrder

    def ReturnCards(self, sender: object):
        """
        Event handler for requesting the return of cards
        """
        self.clear()

    def __str__(self):
        return self.Name + ": " + super().__str__()

    def sort(self, *args, **kwargs):
        if "key" in kwargs.keys():
            return super().sort(*args, **kwargs)
        else:
            return super().sort(*args, key = lambda card: (card.Value + 1) * (card.Suit + 1), **kwargs)


from ._NDice import NDice

class Dice8(NDice):
    """
    Simple dice with 8 sides

    __init__():
        int seed --> seed for the random generator (deafult is None)
    """
    def __init__(self, seed = None):
        super.__init__(8, seed)